from DIRAC import S_OK
from DIRAC.DataManagementSystem.Agent.RequestOperations.RemoveReplica import (
    RemoveReplica,
)
from DIRAC.DataManagementSystem.Client.DataManager import DataManager

from CTADIRAC.DataManagementSystem.Agent.RequestOperations.CTADMSRequestOperationsBase import (
    CTADMSRequestOperationsBase,
)


class CTARemoveReplica(RemoveReplica, CTADMSRequestOperationsBase):
    """CTA specific RemoveReplica class to handle lfns starting with both the old and new VO name."""

    def __call__(self):
        """Remove replicas."""

        self._initMonitoring()
        # # prepare list of targetSEs
        targetSEs = self.operation.targetSEList
        # # check targetSEs for removal
        bannedTargets = self.checkSEsRSS(targetSEs, access="RemoveAccess")
        if not bannedTargets["OK"]:
            self._recordMonitoring(["Attempted", "Failed"], len(self.operation))
            self._commitMonitoring()
            return bannedTargets

        if bannedTargets["Value"]:
            return S_OK(
                f"{','.join(bannedTargets['Value'])} targets are banned for removal"
            )

        # # get waiting files
        waitingFiles = self.getWaitingFilesList()
        # # and prepare dict
        toRemoveDict = {opFile.LFN: opFile for opFile in waitingFiles}
        ## LA: Added for CTA-DIRAC Legacy
        first_lfn = next(iter(toRemoveDict))
        voName = first_lfn.split("/")[1]
        self.dm = DataManager(vo=voName)

        self.log.info(
            f"Todo: {len(toRemoveDict)} replicas to delete from {len(targetSEs)} SEs"
        )

        self._recordMonitoring("Attempted", len(toRemoveDict))

        # # keep status for each targetSE
        removalStatus = {lfn: dict.fromkeys(targetSEs, None) for lfn in toRemoveDict}

        # # loop over targetSEs
        for targetSE in targetSEs:
            self._processTargetSE(targetSE, toRemoveDict, removalStatus)

        # # update file status for waiting files
        failed = self._updateOpFileStatus(removalStatus)

        if failed:
            self.operation.Error = f"failed to remove {failed} replicas"

        self._commitMonitoring()

        return S_OK()

    def _processTargetSE(self, targetSE, toRemoveDict, removalStatus):
        self.log.info(f"Removing replicas at {targetSE}")
        # # 1st step - bulk removal
        bulkRemoval = self._bulkRemoval(toRemoveDict, targetSE)
        if not bulkRemoval["OK"]:
            self.log.error(f"Bulk replica removal failed {bulkRemoval['Message']}")

            self._commitMonitoring()

            return bulkRemoval

        # # report removal status for successful files
        opFiles = [opFile for opFile in toRemoveDict.values() if not opFile.Error]
        self._recordMonitoring("Successful", len(opFiles))

        # # 2nd step - process the rest again
        toRetry = {lfn: opFile for lfn, opFile in toRemoveDict.items() if opFile.Error}
        for lfn, opFile in toRetry.items():
            self._removeWithOwnerProxy(opFile, targetSE)
            if opFile.Error:
                self._recordMonitoring("Failed", 1)
                removalStatus[lfn][targetSE] = opFile.Error
            else:
                self._recordMonitoring("Successful", 1)

    def _updateOpFileStatus(self, removalStatus):
        failed = 0
        for opFile in self.operation:
            if opFile.Status == "Waiting":
                errors = list(
                    {err for err in removalStatus[opFile.LFN].values() if err}
                )
                if errors:
                    opFile.Error = "\n".join(errors)
                    # This seems to be the only unrecoverable error
                    if "Write access not permitted for this credential" in opFile.Error:
                        failed += 1
                        opFile.Status = "Failed"
                else:
                    opFile.Status = "Done"

        return failed



#!/usr/bin/env python

__RCSID__ = "$Id$"

# generic imports
import os
from multiprocessing import Pool

# DIRAC imports
import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script
from DIRAC.DataManagementSystem.Client.DataManager import DataManager

from CTADIRAC.Core.Utilities.tool_box import read_inputs_from_file

Script.setUsageMessage(
    """
Extension of dirac-dms-replicate-lfn in multi-thread mode
Allows for bulk replication of a list of LFNs to a destination Storage Element
Usage:
   cta-prod-replicate-lfn <lfn> <SE>
   cta-prod-replicate-lfn <ascii file with lfn list> <SE>
"""
)

Script.parseCommandLine(ignoreErrors=True)

args = Script.getPositionalArgs()
if len(args) > 1:
    infile = args[0]
    SE = args[1]
else:
    Script.showHelp()

DIRAC.initialize()


@Script()
def main():
    if os.path.isfile(infile):
        infileList = read_inputs_from_file(infile)
    else:
        infileList = [infile]
    p = Pool(10)
    p.map(replicateFile, infileList)


def replicateFile(lfn):
    voName = lfn.split("/")[1]
    if voName not in ["ctao", "vo.cta.in2p3.fr"]:
        message = (
            f"Wrong lfn: path must start with vo name (ctao or vo.cta.in2p3.fr):\n{lfn}"
        )
        gLogger.error(message)
        return
    dm = DataManager(vo=voName)
    res = dm.replicateAndRegister(lfn, SE)
    if not res["OK"]:
        gLogger.error("Error replicating file", lfn)
        return res["Message"]


if __name__ == "__main__":
    main()



from unittest.mock import MagicMock, call, mock_open, patch

import pytest
from ruamel.yaml import YAML

from CTADIRAC.Interfaces.API.CTAJob import (
    MetadataDict,
)
from CTADIRAC.Interfaces.API.CtapipeMergeJob import CtapipeMergeJob
from CTADIRAC.Interfaces.API.CtapipeProcessJob import CtapipeProcessJob
from CTADIRAC.Interfaces.API.MCPipeJob import MCPipeJob
from CTADIRAC.Interfaces.API.MCSimTelProcessJob import MCSimTelProcessJob
from CTADIRAC.Interfaces.API.SimPipeJob import SimPipeJob
from tests.unit.production import (
    COMMON_CONFIG,
    CTAPIPE_PROCESS_METADATA,
    CTAPIPE_PROCESS_OUTPUT_METADATA,
    MERGING1_METADATA,
    MERGING1_OUTPUT_METADATA,
    MERGING_CONFIG_1,
    PROCESSING_CONFIG,
    SIMPIPE_CONFIG,
    SIMPIPE_OUTPUT_METADATA,
    SIMULATION_CONFIG,
    SIMULATION_OUTPUT_METADATA,
)

yaml = YAML(typ="safe", pure=True)
software_version = "v0.19.2"
parents_list: list[int] = [1, 2, 3]

KEY_VALUE_STR = '{"key": "value"}'
TEMP_OUTPUT_DIRECTORY = "/output/directory"


def test_metadata_dict() -> None:
    metadata = MetadataDict()
    key = "unknown_key"
    error_message = f"Key '{key}' is not allowed in MetadataDict"
    with pytest.raises(KeyError) as exc_info:
        metadata[key] = "unknown"
    assert error_message in str(exc_info)


sim_job = MCPipeJob()
process_job = CtapipeProcessJob()
merge_job = CtapipeMergeJob()
sim_process_job = MCSimTelProcessJob()


def test_set_output_metadata() -> None:
    # Simulation Job:
    # Setting class variables 'as' done in WorkflowElement
    sim_job.particle = SIMULATION_CONFIG["job_config"]["particle"]
    sim_job.array_layout = SIMULATION_CONFIG["job_config"]["array_layout"]
    sim_job.site = SIMULATION_CONFIG["job_config"]["site"]
    sim_job.pointing_dir = SIMULATION_CONFIG["job_config"]["pointing_dir"]
    sim_job.version = SIMULATION_CONFIG["job_config"]["version"]
    sim_job.MCCampaign = COMMON_CONFIG["MCCampaign"]
    sim_job.configuration_id = COMMON_CONFIG["configuration_id"]
    assert sim_job.output_metadata == MetadataDict()
    sim_job.set_output_metadata(SIMULATION_OUTPUT_METADATA)
    assert sim_job.output_metadata == SIMULATION_OUTPUT_METADATA

    # Processing Job:
    # Setting class variables 'as' done in WorkflowElement
    process_job.particle = SIMULATION_CONFIG["job_config"]["particle"]
    process_job.array_layout = PROCESSING_CONFIG["job_config"]["array_layout"]
    process_job.site = SIMULATION_CONFIG["job_config"]["site"]
    process_job.pointing_dir = SIMULATION_CONFIG["job_config"]["pointing_dir"]
    process_job.version = PROCESSING_CONFIG["job_config"]["version"]
    process_job.MCCampaign = COMMON_CONFIG["MCCampaign"]
    process_job.configuration_id = COMMON_CONFIG["configuration_id"]

    assert process_job.output_metadata == MetadataDict()
    process_job.set_output_metadata(CTAPIPE_PROCESS_METADATA)
    assert process_job.output_metadata == CTAPIPE_PROCESS_OUTPUT_METADATA

    # Merging Job:
    # Setting class variables 'as' done in WorkflowElement
    merge_job.MCCampaign = COMMON_CONFIG["MCCampaign"]
    merge_job.version = MERGING_CONFIG_1["job_config"]["version"]

    assert merge_job.output_metadata == MetadataDict()
    merge_job.set_output_metadata(MERGING1_METADATA)
    assert merge_job.output_metadata == MERGING1_OUTPUT_METADATA


def test_set_site() -> None:
    sites = ["Paranal", "LaPalma"]
    for site in sites:
        sim_job.set_site(site)
        assert sim_job.site == site


def test_set_particle() -> None:
    particles = ["gamma", "gamma-diffuse", "electron", "proton", "helium"]
    for particle in particles:
        sim_job.set_particle(particle)
        assert sim_job.particle == particle


def test_set_pointing_dir() -> None:
    pointing_dirs: list[str] = ["North", "South", "East", "West"]
    for pointing in pointing_dirs:
        sim_job.set_pointing_dir(pointing)
        assert sim_job.pointing_dir == pointing


def test_set_moon() -> None:
    moon: list[str] = ["dark"]
    sim_job.set_moon(moon)
    assert sim_job.moon == ""
    assert sim_job.output_file_metadata["nsb"] == [1]

    moon = ["dark", "half"]
    sim_job.set_moon(moon)
    assert sim_job.moon == "--with-half-moon"
    assert sim_job.output_file_metadata["nsb"] == [1, 5]

    moon = ["dark", "half", "full"]
    sim_job.set_moon(moon)
    assert sim_job.moon == "--with-full-moon"
    assert sim_job.output_file_metadata["nsb"] == [1, 5, 19]

    moon: str = "dark"
    sim_process_job.set_moon(moon)
    assert sim_process_job.moon == ""
    assert sim_process_job.output_file_metadata["nsb"] == 1

    moon: str = "half"
    sim_process_job.set_moon(moon)
    assert sim_process_job.moon == "--with-half-moon"
    assert sim_process_job.output_file_metadata["nsb"] == 5

    moon: str = "full"
    sim_process_job.set_moon(moon)
    assert sim_process_job.moon == "--with-full-moon"
    assert sim_process_job.output_file_metadata["nsb"] == 19


def test_set_div_ang() -> None:
    div_ang = [
        "0.0098",
        "0.0075",
        "0.0089",
        "0.01568",
        "0.04568",
    ]
    with pytest.raises(SystemExit) as exc_info:
        sim_job.set_div_ang(div_ang)
    assert str(exc_info.value) == "-1"


def test_set_magic() -> None:
    sim_job.set_magic(True)
    assert sim_job.magic == "--with-magic"


def set_sct() -> None:
    sim_version = sim_job.version
    sim_job.set_sct(None)
    assert sim_job.sct == ""
    assert sim_job.version == sim_version

    sim_job.set_sct("all")
    assert sim_job.sct == "--with-all-scts"
    assert sim_job.version == sim_version + "-sc"

    sim_job.set_sct("non-alpha")
    assert sim_job.sct == "--with-sct"
    assert sim_job.version == sim_version + "-sc"


def test_set_output_metadata_simpipe():
    mock_file_content = """
    array_layout_name: Alpha
    site: North
    primary: gamma-diffuse
    zenith_angle: 20.0
    azimuth_angle: North
    model_version: 0.6.0
    """

    with patch("builtins.open", mock_open(read_data=mock_file_content)):
        simpipe_job = SimPipeJob()
        simpipe_job.set_simpipe_config("mock_config.yaml")

        simpipe_job.version = SIMPIPE_CONFIG["job_config"]["version"]
        simpipe_job.MCCampaign = COMMON_CONFIG["MCCampaign"]
        simpipe_job.configuration_id = COMMON_CONFIG["configuration_id"]

        metadata = MetadataDict(
            array_layout=simpipe_job.array_layout,
            site=simpipe_job.site,
            particle=simpipe_job.particle,
            phiP=float(round((float(simpipe_job.azimuth_angle) + 180) % 360, 2)),
            thetaP=float(simpipe_job.zenith_angle),
            sct="True" if simpipe_job.sct else "False",
            outputType=simpipe_job.output_type,
        )

        simpipe_job.set_output_metadata(metadata)

        expected_metadata = MetadataDict(
            array_layout=SIMPIPE_OUTPUT_METADATA["array_layout"],
            site=SIMPIPE_OUTPUT_METADATA["site"],
            particle=SIMPIPE_OUTPUT_METADATA["particle"],
            phiP=SIMPIPE_OUTPUT_METADATA["phiP"],
            thetaP=SIMPIPE_OUTPUT_METADATA["thetaP"],
            sct=SIMPIPE_OUTPUT_METADATA["sct"],
            tel_sim_prog=SIMPIPE_OUTPUT_METADATA["tel_sim_prog"],
            tel_sim_prog_version=SIMPIPE_OUTPUT_METADATA["tel_sim_prog_version"],
            data_level=SIMPIPE_OUTPUT_METADATA["data_level"],
            outputType=SIMPIPE_OUTPUT_METADATA["outputType"],
            configuration_id=SIMPIPE_OUTPUT_METADATA["configuration_id"],
            MCCampaign=SIMPIPE_OUTPUT_METADATA["MCCampaign"],
        )
        assert simpipe_job.output_metadata == expected_metadata


def test_build_file_metadata_with_propagate_run_number():
    simpipe_job = SimPipeJob()
    simpipe_job.run_number = "@{JOB_ID}"
    simpipe_job.run_number_offset = 10
    combination = {"key1": "value1", "key2": "value2"}
    result = simpipe_job.build_file_metadata(combination, propagate_run_number=True)

    # Assert that runNumber is included and equals the default run_number
    assert "runNumber" in result
    assert result["runNumber"] == "@{JOB_ID}"

    # Assert that combination keys and values are included
    assert result["key1"] == "value1"
    assert result["key2"] == "value2"


def test_build_file_metadata_without_propagate_run_number():
    simpipe_job = SimPipeJob()
    simpipe_job.run_number = "@{JOB_ID}"
    simpipe_job.run_number_offset = 10
    combination = {"key1": "value1", "key2": "value2"}
    result = simpipe_job.build_file_metadata(combination, propagate_run_number=False)

    # Assert that runNumber is not included
    assert "runNumber" not in result

    # Assert that combination keys and values are included
    assert result["key1"] == "value1"
    assert result["key2"] == "value2"


def test_build_file_metadata_with_custom_run_number():
    simpipe_job = SimPipeJob()
    simpipe_job.run_number = 1000
    simpipe_job.run_number_offset = 10
    combination = {"key1": "value1"}
    result = simpipe_job.build_file_metadata(combination, propagate_run_number=True)

    # Assert that runNumber includes the offset
    assert "runNumber" in result
    assert result["runNumber"] == 1010  # 1000 + offset (10)

    # Assert that combination keys and values are included
    assert result["key1"] == "value1"


@patch("CTADIRAC.Interfaces.API.SimPipeJob.json.dumps")
@patch("CTADIRAC.Interfaces.API.SimPipeJob.SimPipeJob.setExecutable")
def test_run_simpipe(mock_set_executable, mock_json_dumps):
    job = SimPipeJob()
    job.simpipe_config_options = {"key": "value"}
    job.run_number_offset = 10
    job.run_number = 12345
    mock_json_dumps.return_value = KEY_VALUE_STR

    mock_set_executable.return_value = {
        "Value": {
            "name": "Step_SimPipe",
            "descr_short": "Run SimPipe simulation step",
        }
    }

    job.run_simpipe()

    mock_set_executable.assert_called_once_with(
        "dirac_simpipe_simulate_prod_wrapper",
        arguments='10 12345 \'{"key": "value"}\'',
        logFile="SimPipe_Log.txt",
        modulesList=["cta_script"],
    )
    step = mock_set_executable.return_value
    assert step["Value"]["name"] == "Step_SimPipe"
    assert step["Value"]["descr_short"] == "Run SimPipe simulation step"


def test_set_array_layout_without_scts():
    """Test _set_array_layout when the value does not contain '_scts'."""
    simpipe_job = SimPipeJob()
    simpipe_job._set_array_layout("array_layout_standard")
    assert simpipe_job.array_layout == "array_layout_standard"
    assert not simpipe_job.sct


def test_set_array_layout_with_scts():
    """Test _set_array_layout when the value contains '_scts'."""
    simpipe_job = SimPipeJob()
    simpipe_job._set_array_layout("array_layout_scts")
    assert simpipe_job.array_layout == "array_layout_scts"
    assert simpipe_job.sct


def test_set_site_north():
    """Test _set_site with value 'North'."""
    simpipe_job = SimPipeJob()
    simpipe_job._set_site("North")
    assert simpipe_job.site == "LaPalma"


def test_set_site_south():
    """Test _set_site with value 'South'."""
    simpipe_job = SimPipeJob()
    simpipe_job._set_site("South")
    assert simpipe_job.site == "Paranal"


def test_set_site_invalid():
    """Test _set_site with an invalid value."""
    simpipe_job = SimPipeJob()

    with pytest.raises(SystemExit) as exc_info:
        simpipe_job._set_site("InvalidSite")
    assert str(exc_info.value) == "-1"


def test_azimuth_angle_not_set():
    """Test azimuth_angle property when _azimuth_angle is not set."""
    simpipe_job = SimPipeJob()

    with pytest.raises(SystemExit) as exc_info:
        _ = simpipe_job.azimuth_angle
    assert str(exc_info.value) == "-1"


def test_azimuth_angle_set():
    """Test azimuth_angle property when _azimuth_angle is set."""
    simpipe_job = SimPipeJob()
    simpipe_job._azimuth_angle = 45.0
    assert simpipe_job.azimuth_angle == 45.0


def test_azimuth_angle_set_valid_float():
    """Test azimuth_angle setter with a valid float value."""
    simpipe_job = SimPipeJob()
    simpipe_job.azimuth_angle = 90.0
    assert simpipe_job._azimuth_angle == 90.0


def test_azimuth_angle_set_valid_direction():
    """Test azimuth_angle setter with a valid direction string."""
    simpipe_job = SimPipeJob()
    simpipe_job.azimuth_angle = "north"
    assert simpipe_job._azimuth_angle == 0


def test_azimuth_angle_set_invalid_direction():
    """Test azimuth_angle setter with an invalid direction string."""
    simpipe_job = SimPipeJob()

    with pytest.raises(SystemExit) as exc_info:
        simpipe_job.azimuth_angle = "invalid_direction"
    assert str(exc_info.value) == "-1"


def test_set_simulation_software_corsika():
    """Test _set_simulation_software with value 'corsika'."""
    simpipe_job = SimPipeJob()
    simpipe_job._set_simulation_software("corsika")
    assert simpipe_job.only_corsika
    assert simpipe_job.program_category == "airshower_sim"
    assert simpipe_job.prog_name == "corsika"


def test_set_simulation_software_invalid():
    """Test _set_simulation_software with an invalid value."""
    simpipe_job = SimPipeJob()
    simpipe_job._set_simulation_software("invalid_software")
    assert not simpipe_job.only_corsika
    assert simpipe_job.program_category == "tel_sim"
    assert simpipe_job.prog_name == "simpipe"


@patch("CTADIRAC.Interfaces.API.SimPipeJob.SimPipeJob.setExecutable")
def test_upload_and_register_file_data(mock_set_executable):
    """Test upload_and_register_file with data type."""
    simpipe_job = SimPipeJob()
    simpipe_job.output_data_type = "data"
    simpipe_job.package = "simpipe"
    simpipe_job.program_category = "tel_sim"
    simpipe_job.catalogs = "catalog"

    meta_data_json = KEY_VALUE_STR
    file_meta_data_json = '{"file_key": "file_value"}'
    data_output_pattern = "/output/pattern/*.zst"
    log_str = "test"

    mock_set_executable.return_value = {
        "Value": {
            "name": "Step_DataManagement",
            "descr_short": "Save data files to SE and register them in DFC",
        }
    }

    simpipe_job.upload_and_register_file(
        meta_data_json,
        file_meta_data_json,
        data_output_pattern,
        log_str,
        data_type="data",
    )

    mock_set_executable.assert_called_once_with(
        "cta-prod-managedata",
        arguments=(
            f"'{meta_data_json}' '{file_meta_data_json}' "
            f"{simpipe_job.base_path} '{data_output_pattern}' "
            f"{simpipe_job.package} {simpipe_job.program_category} "
            f"'{simpipe_job.catalogs}' {simpipe_job.output_data_type}"
        ),
        logFile=f"DataManagement_{log_str}_Log.txt",
    )
    step = mock_set_executable.return_value
    assert step["Value"]["name"] == "Step_DataManagement"
    assert (
        step["Value"]["descr_short"] == "Save data files to SE and register them in DFC"
    )


@patch("CTADIRAC.Interfaces.API.SimPipeJob.SimPipeJob.setExecutable")
def test_upload_and_register_file_log(mock_set_executable):
    """Test upload_and_register_file with log type."""
    simpipe_job = SimPipeJob()
    simpipe_job.output_log_type = "log"
    simpipe_job.package = "simpipe"
    simpipe_job.program_category = "tel_sim"
    simpipe_job.catalogs = "catalog"

    meta_data_json = KEY_VALUE_STR
    file_meta_data_json = '{"file_key": "file_value"}'
    log_file_pattern = "/output/pattern/*.gz"
    log_str = "test"

    mock_set_executable.return_value = {
        "Value": {
            "name": "Step_LogManagement",
            "descr_short": "Save log files to SE and register them in DFC",
        }
    }

    simpipe_job.upload_and_register_file(
        meta_data_json, file_meta_data_json, log_file_pattern, log_str, data_type="log"
    )

    mock_set_executable.assert_called_once_with(
        "cta-prod-managedata",
        arguments=(
            f"'{meta_data_json}' '{file_meta_data_json}' "
            f"{simpipe_job.base_path} '{log_file_pattern}' "
            f"{simpipe_job.package} {simpipe_job.program_category} "
            f"'{simpipe_job.catalogs}' {simpipe_job.output_log_type}"
        ),
        logFile=f"LogManagement_{log_str}_Log.txt",
    )
    step = mock_set_executable.return_value
    assert step["Value"]["name"] == "Step_LogManagement"
    assert (
        step["Value"]["descr_short"] == "Save log files to SE and register them in DFC"
    )


@patch("CTADIRAC.Interfaces.API.SimPipeJob.SimPipeJob.setExecutable")
def test_upload_and_register_file_reduced_event_lists(mock_set_executable):
    """Test upload_and_register_file with reduced_event_lists type."""
    simpipe_job = SimPipeJob()
    simpipe_job.output_reduced_event_lists_type = "reduced_event_lists"
    simpipe_job.package = "simpipe"
    simpipe_job.program_category = "tel_sim"
    simpipe_job.catalogs = "catalog"

    meta_data_json = KEY_VALUE_STR
    file_meta_data_json = '{"file_key": "file_value"}'
    reduced_event_pattern = "/output/pattern/*.reduced_event_data.hdf5"
    log_str = "test"

    mock_set_executable.return_value = {
        "Value": {
            "name": "Step_ReducedEventListsManagement",
            "descr_short": "Save reduced event lists to SE and register them in DFC",
        }
    }

    simpipe_job.upload_and_register_file(
        meta_data_json,
        file_meta_data_json,
        reduced_event_pattern,
        log_str,
        data_type="reduced_event_lists",
    )

    mock_set_executable.assert_called_once_with(
        "cta-prod-managedata",
        arguments=(
            f"'{meta_data_json}' '{file_meta_data_json}' "
            f"{simpipe_job.base_path} '{reduced_event_pattern}' "
            f"{simpipe_job.package} {simpipe_job.program_category} "
            f"'{simpipe_job.catalogs}' {simpipe_job.output_reduced_event_lists_type}"
        ),
        logFile=f"ReducedEventListsManagement_{log_str}_Log.txt",
    )
    step = mock_set_executable.return_value
    assert step["Value"]["name"] == "Step_ReducedEventListsManagement"
    assert (
        step["Value"]["descr_short"]
        == "Save reduced event lists to SE and register them in DFC"
    )


@patch("CTADIRAC.Interfaces.API.SimPipeJob.SimPipeJob.setExecutable")
def test_upload_and_register_file_invalid_type(mock_set_executable):
    """Test upload_and_register_file with invalid data type."""
    simpipe_job = SimPipeJob()

    meta_data_json = KEY_VALUE_STR
    file_meta_data_json = '{"file_key": "file_value"}'
    file_pattern = "/output/pattern/*"
    log_str = "test"

    with pytest.raises(ValueError) as exc_info:
        simpipe_job.upload_and_register_file(
            meta_data_json,
            file_meta_data_json,
            file_pattern,
            log_str,
            data_type="invalid_type",
        )

    assert "Unknown data type: invalid_type" in str(exc_info.value)
    expected_types = "Supported types: ['data', 'log', 'reduced_event_lists']"
    assert expected_types in str(exc_info.value)


def test_run_dedicated_software_calls_run_simpipe():
    """Test run_dedicated_software to ensure it calls run_simpipe."""
    simpipe_job = SimPipeJob()

    # Mock the run_simpipe method
    simpipe_job.run_simpipe = lambda: setattr(simpipe_job, "run_simpipe_called", True)
    simpipe_job.run_simpipe_called = False

    simpipe_job.run_dedicated_software()

    assert (
        simpipe_job.run_simpipe_called
    ), "run_simpipe was not called by run_dedicated_software"


def test_run_dedicated_software_no_side_effects():
    """Test run_dedicated_software to ensure no unexpected side effects."""
    simpipe_job = SimPipeJob()

    # Mock the run_simpipe method
    simpipe_job.run_simpipe = lambda: None

    # Capture initial state
    initial_state = simpipe_job.__dict__.copy()

    simpipe_job.run_dedicated_software()

    # Ensure no state changes occurred
    assert (
        simpipe_job.__dict__ == initial_state
    ), "Unexpected side effects in run_dedicated_software"


def test_set_metadata_and_register_data():
    """Test for set_metadata_and_register_data with valid inputs."""
    simpipe_job = SimPipeJob()
    simpipe_job.output_metadata = {"key": "value"}
    simpipe_job.model_version = ["v1", "v2"]
    simpipe_job.output_directory = TEMP_OUTPUT_DIRECTORY
    simpipe_job.only_corsika = False

    simpipe_job.build_file_metadata = MagicMock(return_value={"runNumber": "1234"})
    simpipe_job.upload_and_register_file = MagicMock()

    simpipe_job.set_metadata_and_register_data()

    expected_calls = [
        call({"model_version": "v1"}, True),
        call({"model_version": "v2"}, True),
    ]
    simpipe_job.build_file_metadata.assert_has_calls(expected_calls, any_order=False)

    # Should call upload_and_register_file 4 times total (2 data + 2 log files)
    assert simpipe_job.upload_and_register_file.call_count == 4


def test_set_metadata_and_register_data_with_corsika():
    """Test for set_metadata_and_register_data when only_corsika is True."""
    simpipe_job = SimPipeJob()
    simpipe_job.output_metadata = {"key": "value"}
    simpipe_job.model_version = ["v1"]
    simpipe_job.output_directory = TEMP_OUTPUT_DIRECTORY
    simpipe_job.only_corsika = True

    simpipe_job.build_file_metadata = MagicMock(return_value={"runNumber": "1234"})
    simpipe_job.upload_and_register_file = MagicMock()

    simpipe_job.set_metadata_and_register_data()

    expected_calls = [
        call({"model_version": "v1"}, True),
    ]
    simpipe_job.build_file_metadata.assert_has_calls(expected_calls, any_order=False)

    # Should call upload_and_register_file 2 times (1 data + 1 log file)
    assert simpipe_job.upload_and_register_file.call_count == 2


def test_set_metadata_and_register_data_empty_model_version():
    """Test for set_metadata_and_register_data with empty model_version."""
    simpipe_job = SimPipeJob()
    simpipe_job.output_metadata = {"key": "value"}
    simpipe_job.model_version = []  # Empty model_version
    simpipe_job.output_directory = TEMP_OUTPUT_DIRECTORY
    simpipe_job.only_corsika = False

    simpipe_job.build_file_metadata = MagicMock()
    simpipe_job.upload_and_register_file = MagicMock()

    simpipe_job.set_metadata_and_register_data()

    simpipe_job.build_file_metadata.assert_not_called()
    simpipe_job.upload_and_register_file.assert_not_called()


def test_set_metadata_and_register_data_with_reduced_event_lists():
    """Test set_metadata_and_register_data with reduced event lists enabled."""
    simpipe_job = SimPipeJob()
    simpipe_job.output_metadata = {"key": "value"}
    simpipe_job.model_version = ["v1"]
    simpipe_job.output_directory = TEMP_OUTPUT_DIRECTORY
    simpipe_job.only_corsika = False
    simpipe_job.save_reduced_event_lists = True

    simpipe_job.build_file_metadata = MagicMock(return_value={"runNumber": "1234"})
    simpipe_job.upload_and_register_file = MagicMock()

    simpipe_job.set_metadata_and_register_data()

    expected_calls = [
        call({"model_version": "v1"}, True),
    ]
    simpipe_job.build_file_metadata.assert_has_calls(expected_calls, any_order=False)

    # Should call upload_and_register_file 3 times:
    # 1 data + 1 reduced_event_lists + 1 log file
    assert simpipe_job.upload_and_register_file.call_count == 3

    # Verify the calls include the reduced event lists
    upload_calls = simpipe_job.upload_and_register_file.call_args_list
    data_types_called = [call[1]["data_type"] for call in upload_calls]
    assert "data" in data_types_called
    assert "reduced_event_lists" in data_types_called
    assert "log" in data_types_called


def test_set_systematic_uncertainty_to_test() -> None:
    systematic_uncertainty_to_test: str = "LaPalma/clouds/ID1"
    sim_process_job.set_systematic_uncertainty_to_test(systematic_uncertainty_to_test)
    assert (
        sim_process_job.systematic_uncertainty_to_test == systematic_uncertainty_to_test
    )


def test_run_sim_telarray(mocker):
    systematic_uncertainty_to_test: str = "LaPalma/clouds/ID1"
    sim_process_job.set_systematic_uncertainty_to_test(systematic_uncertainty_to_test)

    mock_step = {"Value": {"name": "", "descr_short": ""}}
    mock_set_exec = mocker.patch.object(
        sim_process_job, "setExecutable", return_value=mock_step
    )

    sim_process_job.run_sim_telarray(debug=False)

    mock_set_exec.assert_called_once_with(
        "./dirac_sim_telarray_process",
        arguments="LaPalma/clouds/ID1",
        logFile="Simtel_Log.txt",
    )

    assert mock_step["Value"]["name"] == "Step_Simtel"
    assert (
        mock_step["Value"]["descr_short"]
        == "Run sim_telarray processing of CORSIKA file"
    )



import pytest
from cwl_utils.parser.cwl_v1_2 import (
    CommandLineTool,
    CommandInputParameter,
    CommandOutputParameter,
    CommandInputArraySchema,
    CommandOutputArraySchema,
    File,
    InputArraySchema,
    OutputArraySchema,
    Workflow,
    WorkflowStep,
)
from cwl_utils.parser import save

from CTADIRAC.Interfaces.Utilities.CWLUtilities import (
    format_input_by_type,
    format_output_by_type,
    get_current_step_obj,
    get_step_input_type,
    get_step_output_type,
    set_basename_value,
    set_input_file_basename,
    verify_cwl_output_type,
    fill_defaults,
)

ARRAY_FILE_OUTPUT = CommandOutputArraySchema(items="test.txt", type_="File")
ARRAY_ARRAY_OUTPUT = CommandOutputArraySchema(
    items=["test.txt", "test2.txt"], type_="array"
)


@pytest.mark.parametrize(
    ("output_type", "expected_result"),
    [
        ("File", True),
        (ARRAY_FILE_OUTPUT, True),
        (ARRAY_ARRAY_OUTPUT, True),
        (["File"], True),
        (["null", "File"], True),
        (["null", ARRAY_FILE_OUTPUT], True),
        (["null", ARRAY_ARRAY_OUTPUT], True),
        ("string", False),
        (["null", "string"], False),
    ],
)
def test_verify_cwl_output_type(output_type, expected_result):
    result = verify_cwl_output_type(output_type)
    assert result is expected_result


@pytest.mark.parametrize(
    ("cwl", "inputs", "expected_inputs"),
    [
        (
            CommandLineTool(
                inputs=[
                    CommandInputParameter(
                        id="/some/path#step_1",
                        type_="string",
                    )
                ],
                outputs={},
            ),
            {"input1": "input1", "input2": "input2"},
            {"input1": "input1", "input2": "input2"},
        ),
        (
            Workflow(
                steps=[
                    WorkflowStep(
                        id="/some/path#step_1",
                        in_=[],
                        out=[],
                        run=CommandLineTool(
                            inputs=[
                                CommandInputParameter(
                                    id="/some/path#step_1",
                                    type_="string",
                                    default="defaultInputStep1",
                                )
                            ],
                            outputs=[],
                            baseCommand="echo CLT1",
                        ),
                    ),
                    WorkflowStep(
                        id="/some/path#step_2",
                        in_=[],
                        out=[],
                        run=CommandLineTool(
                            inputs=[
                                CommandInputParameter(
                                    id="/some/path#step_2",
                                    type_="string",
                                    default="defaultInputStep2",
                                )
                            ],
                            outputs=[],
                            baseCommand="echo CLT2",
                        ),
                    ),
                ],
                inputs={},
                outputs={},
            ),
            {"input1": "input1", "input2": "input2"},
            {
                "input1": "input1",
                "input2": "input2",
                "step_1": "defaultInputStep1",
                "step_2": "defaultInputStep2",
            },
        ),
    ],
)
def test_fill_defaults(cwl, inputs, expected_inputs):
    result = fill_defaults(cwl, inputs)
    assert result == expected_inputs


@pytest.mark.parametrize(
    ("input_file", "expected_basename"),
    [
        (save({"input1": File(path="test_lfn_file.txt")}), "test_lfn_file.txt"),
        ({"input1": File(path="test_lfn_file.txt")}, "test_lfn_file.txt"),
        (
            save({"input1": File(path="test_local_file.txt", basename="local_file")}),
            "local_file",
        ),
        (
            {"input1": File(path="test_local_file.txt", basename="local_file")},
            "local_file",
        ),
        ({"input1": {"class": "File", "path": "/some/path/to/file.txt"}}, "file.txt"),
    ],
)
def test_set_input_file_basename(input_file, expected_basename):
    set_input_file_basename(input_file)
    for inp in input_file.values():
        if isinstance(inp, File):
            assert inp.basename == expected_basename
        else:
            assert inp.get("basename") == expected_basename


@pytest.mark.parametrize(
    ("cwl", "step_name", "expected_step_id"),
    [
        (
            Workflow(
                steps=[
                    WorkflowStep(
                        id="/some/path#step_1",
                        in_=[],
                        out=[],
                        run=CommandLineTool(
                            inputs=[
                                CommandInputParameter(
                                    id="/some/path#step_1",
                                    type_="string",
                                    default="defaultInputStep1",
                                )
                            ],
                            outputs=[],
                            baseCommand="echo CLT1",
                        ),
                    ),
                    WorkflowStep(
                        id="/some/path#step_2",
                        in_=[],
                        out=[],
                        run=CommandLineTool(
                            inputs=[
                                CommandInputParameter(
                                    id="/some/path#step_2",
                                    type_="string",
                                    default="defaultInputStep2",
                                )
                            ],
                            outputs=[],
                            baseCommand="echo CLT2",
                        ),
                    ),
                ],
                inputs={},
                outputs={},
            ),
            "step_2",
            "/some/path#step_2",
        ),
    ],
)
def test_get_current_step_obj(cwl, step_name, expected_step_id):
    step = get_current_step_obj(cwl, step_name)
    assert step.id == expected_step_id


@pytest.mark.parametrize(
    ("input", "expected"),
    [
        (
            {"class": "File", "path": "/some/path/to/file.txt"},
            {"class": "File", "path": "/some/path/to/file.txt", "basename": "file.txt"},
        ),
    ],
)
def test_set_basename_value(input, expected):
    result = set_basename_value(input)
    assert result == expected


@pytest.mark.parametrize(
    ("step", "input_name", "expected"),
    [
        (
            WorkflowStep(
                id="/some/path#step_1",
                in_=[],
                out=[],
                run=CommandLineTool(
                    inputs=[
                        CommandInputParameter(
                            id="/some/path#step_1/input1",
                            type_="string",
                            default="defaultInputStep1",
                        ),
                        CommandInputParameter(
                            id="/some/path#step_1/input2",
                            type_="File",
                            default="defaultInputStep2",
                        ),
                    ],
                    outputs=[],
                    baseCommand="echo CLT1",
                ),
            ),
            "input2",
            "File",
        ),
    ],
)
def test_get_step_input_type(step, input_name, expected):
    result = get_step_input_type(step, input_name)
    assert result == expected


@pytest.mark.parametrize(
    ("step", "output_name", "expected"),
    [
        (
            WorkflowStep(
                id="/some/path#step_1",
                in_=[],
                out=[],
                run=CommandLineTool(
                    inputs=[],
                    outputs=[
                        CommandOutputParameter(
                            id="/some/path#step_1/output1",
                            type_="string",
                        ),
                        CommandOutputParameter(
                            id="/some/path#step_1/output2",
                            type_="File",
                        ),
                    ],
                    baseCommand="echo CLT1",
                ),
            ),
            "output2",
            "File",
        ),
    ],
)
def test_get_step_output_type(step, output_name, expected):
    result = get_step_output_type(step, output_name)
    assert result == expected


@pytest.mark.parametrize(
    ("input", "in_type", "expected"),
    [
        # Basic string types
        ("value", "string", "value"),
        ("true", "boolean", "true"),
        ("42", "int", "42"),
        # File type
        ("myfile.txt", "File", {"class": "File", "path": "myfile.txt"}),
        # List types
        ("myfile.txt", ["File", "string"], {"class": "File", "path": "myfile.txt"}),
        ("value", ["string", "null"], "value"),
        ("value", ["null", "string"], "value"),
        ("value", ["null"], "value"),
        # CommandInputArraySchema
        (
            "myfile.txt",
            CommandInputArraySchema(items="myfile.txt", type_="File"),
            {"class": "File", "path": "myfile.txt"},
        ),
        ("value", CommandInputArraySchema(items="value", type_="string"), "value"),
        # InputArraySchema
        (
            "myfile.txt",
            InputArraySchema(items="myfile.txt", type_="File"),
            {"class": "File", "path": "myfile.txt"},
        ),
        ("value", InputArraySchema(items="value", type_="string"), "value"),
        # Edge case: empty list
        ("value", [], "value"),
    ],
)
def test_format_input_by_type(input, in_type, expected):
    result = format_input_by_type(input, in_type)
    assert result == expected


@pytest.mark.parametrize(
    ("output", "out_type", "expected"),
    [
        # Basic string types
        ("value", "string", "value"),
        ("true", "boolean", "true"),
        ("42", "int", "42"),
        # File type
        ("myfile.txt", "File", {"class": "File", "path": "myfile.txt"}),
        # OutputArraySchema
        (
            "myfile.txt",
            OutputArraySchema(items="myfile.txt", type_="File"),
            {"class": "File", "path": "myfile.txt"},
        ),
        ("value", OutputArraySchema(items="value", type_="string"), "value"),
        # Edge case: non-string output
        (None, "File", None),
        # Edge case: empty string output
        ("", "string", ""),
    ],
)
def test_format_output_by_type(output, out_type, expected):
    result = format_output_by_type(output, out_type)
    assert result == expected



from DIRAC.MonitoringSystem.Client.MonitoringReporter import MonitoringReporter
from DIRAC.DataManagementSystem.Agent.RequestOperations.DMSRequestOperationsBase import (
    DMSRequestOperationsBase,
)


class CTADMSRequestOperationsBase(DMSRequestOperationsBase):
    def __init__(self, operation=None, csPath=None):
        super().__init__(operation, csPath)
        self.rmsMonitoringReporter = None

    def _initMonitoring(self):
        # The flag  'rmsMonitoring' is set by the RequestTask and is False by default.
        # Here we use 'createRMSRecord' to create the ES record which is defined inside OperationHandlerBase.
        if self.rmsMonitoring:
            self.rmsMonitoringReporter = MonitoringReporter(
                monitoringType="RMSMonitoring"
            )

    def _recordMonitoring(self, status, count):
        if self.rmsMonitoring:
            if not isinstance(status, list):
                status = [status]
            for s in status:
                self.rmsMonitoringReporter.addRecord(self.createRMSRecord(s, count))

    def _commitMonitoring(self):
        if self.rmsMonitoring:
            self.rmsMonitoringReporter.commit()



import json
from unittest.mock import MagicMock

import pytest
from CTADIRAC.Core.Workflow.Modules.ProdDataManager import ProdDataManager

prod_dm = ProdDataManager()


def test_getRunPath():
    filemetadata = {"runNumber": "00000044"}
    res = prod_dm._getRunPath(filemetadata)
    assert res == "000xxx"


@pytest.fixture
def manager():
    mgr = ProdDataManager(catalogs=[])
    # replace the real fc/fcc with mocks
    mgr.fc = MagicMock()
    mgr.fcc = MagicMock()
    return mgr


def test_createMDTopDirectory(manager):
    manager.fc.createDirectory.return_value = {"OK": True}
    manager.fcc.setMetadata.return_value = {"OK": True}
    output_metadata = {
        "MCCampaign": "PROD6",
        "site": "Paranal",
        "particle": "gamma",
        "tel_sim_prog": "sim_telarray",
    }
    metadata_json = json.dumps(output_metadata)
    base_path = "/ctao/test/MC"
    program_category = "tel_sim"
    res = manager.createMDTopDirectory(metadata_json, base_path, program_category)
    path = "/ctao/test/MC/PROD6/Paranal/gamma/sim_telarray"
    assert res["OK"]
    assert res["Value"] == path


def test_createMDSubDirectory(manager):
    manager.fc.createDirectory.return_value = {"OK": True}
    manager.fcc.setMetadata.return_value = {"OK": True}
    output_metadata = {
        "MCCampaign": "PROD6",
        "site": "Paranal",
        "particle": "gamma",
        "tel_sim_prog": "sim_telarray",
        "data_level": -1,
        "configuration_id": 10,
    }
    metadata_json = json.dumps(output_metadata)
    base_path = "/ctao/test/MC/PROD6/Paranal/gamma/sim_telarray"
    program_category = "tel_sim"
    output_type_list = ["Data", "Log"]
    trans_id = 2
    res = manager.createMDSubDirectory(
        metadata_json, base_path, program_category, output_type_list, trans_id
    )
    path = "/ctao/test/MC/PROD6/Paranal/gamma/sim_telarray/2"
    assert res["OK"]
    assert res["Value"] == path, f"Expected: {path}, Got: {res['Value']}"



""" CTA specific ReplicateAndRegister class to handle lfns starting with both the old and new VO name """

# # imports
import re
from collections import defaultdict

# # imports from DIRAC
from DIRAC import S_ERROR, S_OK, gLogger
from DIRAC.Core.Utilities.Adler import compareAdler, hexAdlerToInt, intAdlerToHex
from DIRAC.DataManagementSystem.Agent.RequestOperations.ReplicateAndRegister import (
    ReplicateAndRegister,
)
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Resources.Storage.StorageElement import StorageElement


def filterReplicas(opFile, logger=None, opSources=None, activeReplicas=None):
    """filter out banned/invalid source SEs

    :param list opSources: list of SE names to which limit the possible sources
    :param dict activeReplicas: the result of dm.getActiveReplicas(*)["Value"]. Used as a cache

    :returns: Valid list of SEs valid as source

    """
    ## LA: Added for CTADIRAC Legacy
    voName = opFile.LFN.split("/")[1]
    dataManager = DataManager(vo=voName)
    if logger is None:
        logger = gLogger

    log = logger.getSubLogger("filterReplicas")
    result = defaultdict(list)

    activeReplicas = _getActiveReplicas(opFile, dataManager, activeReplicas, log)

    _checkReplicaExists(opFile, activeReplicas)

    replicas = activeReplicas["Successful"].get(opFile.LFN, {})

    # If user set sourceSEs, only consider those replicas
    if opSources:
        replicas = {x: y for (x, y) in replicas.items() if x in opSources}

    noReplicas = False
    if not replicas:
        noReplicas = _handleNoReplicas(dataManager, opFile, result, log)
        if isinstance(noReplicas, dict):  # Means an error was returned
            return noReplicas

    _setFileChecksumIfMissing(opFile)
    # If no replica was found, return what we collected as information
    if not replicas:
        return S_OK(result)

    for repSEName in replicas:
        ## LA: Added for CTA-DIRAC Legacy
        _processReplica(opFile, voName, repSEName, result, log, noReplicas)

    return S_OK(result)


def _getActiveReplicas(opFile, dataManager, activeReplicas, log):
    if activeReplicas:
        return activeReplicas

    res = dataManager.getActiveReplicas(opFile.LFN, getUrl=False, preferDisk=True)
    if not res["OK"]:
        log.error(f"Failed to get active replicas {res['Message']}")
        return res

    return res["Value"]


def _checkReplicaExists(opFile, activeReplicas):
    reNotExists = re.compile(r".*such file.*")
    failed = activeReplicas["Failed"].get(opFile.LFN, "")
    if reNotExists.match(failed.lower()):
        opFile.Status = "Failed"
        opFile.Error = failed
        return S_ERROR(failed)


def _handleNoReplicas(dataManager, opFile, result, log):
    allReplicas = dataManager.getReplicas(opFile.LFN, getUrl=False)
    if not allReplicas["OK"]:
        return allReplicas

    allReplicas = allReplicas["Value"]["Successful"].get(opFile.LFN, {})
    if not allReplicas:
        result["NoReplicas"].append(None)
        noReplicas = True
    else:
        # There are replicas but we cannot get metadata because the replica is not active
        result["NoActiveReplicas"] += list(allReplicas)
        noReplicas = False
    log.verbose(
        f"File has no{'' if noReplicas else ' active'} replica in File Catalog",
        opFile.LFN,
    )
    return noReplicas


def _setFileChecksumIfMissing(opFile):
    if not opFile.Checksum or hexAdlerToInt(opFile.Checksum) is False:
        # Set Checksum to FC checksum if not set in the request
        fcMetadata = FileCatalog().getFileMetadata(opFile.LFN)
        fcChecksum = (
            fcMetadata.get("Value", {})
            .get("Successful", {})
            .get(opFile.LFN, {})
            .get("Checksum")
        )
        # Replace opFile.Checksum if it doesn't match a valid FC checksum
        if fcChecksum:
            if hexAdlerToInt(fcChecksum) is not False:
                opFile.Checksum = fcChecksum
                opFile.ChecksumType = fcMetadata["Value"]["Successful"][opFile.LFN].get(
                    "ChecksumType", "Adler32"
                )
            else:
                opFile.Checksum = None


def _processReplica(opFile, voName, repSEName, result, log, noReplicas):
    ## LA: Added for CTA-DIRAC Legacy
    repSEMetadata = StorageElement(repSEName, vo=voName).getFileMetadata(opFile.LFN)
    error = repSEMetadata.get(
        "Message", repSEMetadata.get("Value", {}).get("Failed", {}).get(opFile.LFN)
    )
    if error:
        log.warn(
            f"unable to get metadata at {repSEName} for {opFile.LFN}",
            error.replace("\n", ""),
        )
        if "File does not exist" in error or "No such file" in error:
            result["NoReplicas"].append(repSEName)
        else:
            result["NoMetadata"].append(repSEName)
    elif not noReplicas:
        repSEMetadata = repSEMetadata["Value"]["Successful"][opFile.LFN]
        seChecksum = repSEMetadata.get("Checksum")

        ## LA: Bug fix for empty checksum
        if seChecksum == "":
            seChecksum = hexAdlerToInt(False)
        else:
            seChecksum = hexAdlerToInt(seChecksum)
        # As from here seChecksum is an integer or False, not a hex string!
        if seChecksum is False and opFile.Checksum:
            result["NoMetadata"].append(repSEName)
        elif not seChecksum and opFile.Checksum:
            opFile.Checksum = None
            opFile.ChecksumType = None
        elif seChecksum and (not opFile.Checksum or opFile.Checksum == "False"):
            # Use the SE checksum (convert to hex) and force type to be Adler32
            opFile.Checksum = intAdlerToHex(seChecksum)
            opFile.ChecksumType = "Adler32"
        if (
            not opFile.Checksum
            or not seChecksum
            or compareAdler(intAdlerToHex(seChecksum), opFile.Checksum)
        ):
            # # All checksums are OK
            result["Valid"].append(repSEName)
        else:
            log.warn(
                " %s checksum mismatch, FC: '%s' @%s: '%s'"
                % (
                    opFile.LFN,
                    opFile.Checksum,
                    repSEName,
                    intAdlerToHex(seChecksum),
                )
            )
            result["Bad"].append(repSEName)
    else:
        # If a replica was found somewhere, don't set the file as no replicas
        result["NoReplicas"] = []


class CTAReplicateAndRegister(ReplicateAndRegister):
    def _filterReplicas(self, opFile, activeReplicas):
        """filter out banned/invalid source SEs"""
        return filterReplicas(
            opFile,
            logger=self.log,
            opSources=self.operation.sourceSEList,
            activeReplicas=activeReplicas,
        )



from copy import deepcopy
from pathlib import Path
from typing import Any

from cwl_utils.pack import pack
from cwl_utils.parser import load_document, WorkflowStep, save, load_document_by_uri
from cwl_utils.parser import (
    CommandLineTool,
    CommandOutputParameter,
    ExpressionTool,
    File,
    Workflow,
    WorkflowStepInput,
)
from cwl_utils.parser.utils import load_inputfile
from cwl_utils.expression import do_eval

from CTADIRAC.Interfaces.Utilities.CWLUtilities import (
    fill_defaults,
    format_input_by_type,
    format_output_by_type,
    get_step_input_type,
    get_step_output_type,
    set_basename_value,
    verify_cwl_output_type,
    LFN_PREFIX,
    LOCAL_PREFIX,
    LFN_DIRAC_PREFIX,
    JS_REQ,
    set_input_file_basename,
)


class CWLTranslator:
    """Translator from CWL Workflow to DIRAC Job.
    Extract needed DIRAC job arguments from the CWL and inputs description.

    Parameters
    ----------
        cwl_workflow : str
            Path to the local CWL workflow file
        cwl_inputs : str
            Path to the local CWL inputs file
    """

    def __init__(
        self,
        cwl_workflow: str,
        cwl_inputs: str,
    ) -> None:
        self.cwl_workflow_path = Path(cwl_workflow)
        self.cwl_inputs_path = Path(cwl_inputs)

        self.original_cwl = load_document(pack(str(self.cwl_workflow_path)))
        self.unpacked_cwl = load_document_by_uri(self.cwl_workflow_path)
        self.original_inputs = load_inputfile(
            self.original_cwl.cwlVersion, self.cwl_inputs_path.read_text()
        )

        self.transformed_inputs = deepcopy(self.original_inputs)
        self.transformed_cwl = deepcopy(self.original_cwl)
        self.output_sandbox = []
        self.output_data = []
        self.input_sandbox = []
        self.input_data = []

        self.current_step_name = None

    def translate(self, cvmfs_base_path: Path, apptainer_options: list[Any]) -> None:
        """Translate the CWL workflow description into Dirac compliant execution.

        Parameters
        ----------
            cvmfs_base_path : Path
                The base path for CVMFS container repository.
            apptainer_options : list[Any]
                A list of options for Apptainer.
        """

        if isinstance(self.transformed_cwl, CommandLineTool):
            self._translate_clt(cvmfs_base_path, apptainer_options)

        if isinstance(self.transformed_cwl, Workflow):
            self._translate_workflow(cvmfs_base_path, apptainer_options)

    def _translate_clt(
        self, cvmfs_base_path: Path, apptainer_options: list[Any]
    ) -> None:
        """Translate the CWL CommandLineTool description into Dirac compliant execution.

        Parameters
        ----------
            cvmfs_base_path : Path
                The base path for CVMFS container repository.
            apptainer_options : list[Any]
                A list of options for Apptainer.
        """
        self._extract_input_files(self.original_inputs)
        if self.transformed_cwl.hints:
            self.transformed_cwl = self._translate_docker_hints(
                self.transformed_cwl, cvmfs_base_path, apptainer_options
            )

        self._extract_output_files(self.transformed_cwl, self.original_inputs)

    def _translate_workflow(
        self, cvmfs_base_path: Path, apptainer_options: list[Any]
    ) -> None:
        """Translate the CWL Workflow description into Dirac compliant execution.

        Parameters
        ----------
        cvmfs_base_path : Path
            The base path for CVMFS container repository.
        apptainer_options : list[Any]
            A list of options for Apptainer.
        """
        # Extract the DIRAC related
        self._extract_input_files(self.original_inputs)
        # Need to set the file basename for JSReq:
        steps_inputs = deepcopy(self.transformed_inputs)
        self.evaluated_steps_outputs = {}
        for n, wf_step in enumerate(self.transformed_cwl.steps):
            # TODO: handle the case where the step is also a workflow
            if isinstance(wf_step.run, (CommandLineTool, ExpressionTool)):
                step_name = wf_step.id.rpartition("#")[2].split("/")[0]
                self.current_step_name = step_name
                self.evaluated_steps_outputs.setdefault(step_name, {})
                # replace docker hints
                if wf_step.run.hints:
                    self.transformed_cwl.steps[n].run = self._translate_docker_hints(
                        wf_step.run, cvmfs_base_path, apptainer_options
                    )
                # Update the input context for next steps
                steps_inputs = fill_defaults(wf_step.run, steps_inputs)
                steps_inputs = set_input_file_basename(steps_inputs)
                steps_inputs = self._evaluate_step_inputs(wf_step, steps_inputs)

                # Evaluate step outputs
                # specific evaluation if the step is an ExpressionTool
                if isinstance(wf_step.run, ExpressionTool):
                    exptool_outputs = self._evaluate_expression_tool_outputs(
                        self.unpacked_cwl, wf_step, steps_inputs, step_name
                    )
                    self.evaluated_steps_outputs[step_name] = exptool_outputs
                else:
                    self._evaluate_step_outputs(wf_step, steps_inputs)
            elif isinstance(wf_step, Workflow):
                raise NotImplementedError("Workflow steps are not handle yet...")

        # Finally extract the DIRAC related outputs
        self._extract_workflow_outputs()

    @staticmethod
    def _translate_docker_hints(
        cwl_object: CommandLineTool, cvmfs_base_path: Path, apptainer_options: list[Any]
    ) -> CommandLineTool:
        """Translate CWL DockerRequirement into Dirac compliant execution.

        Parameters
        ----------
            cwl_object : CommandLineTool
                The CWL definition.
            cvmfs_base_path : Path
                The base path for CVMFS container repository.
            apptainer_options : list[Any]
                A list of options for Apptainer.

        Returns
        -------
            cwl_object :CommandLineTool
                The translated cwl object.
        """
        for index, hint in enumerate(cwl_object.hints):
            if hint.class_ == "DockerRequirement":
                image = hint.dockerPull
                image_path = str(cvmfs_base_path / f"{image}")

                cmd = [
                    "apptainer",
                    "run",
                    *apptainer_options,
                    image_path,
                ]

                if isinstance(cwl_object.baseCommand, str):
                    cmd.append(cwl_object.baseCommand)
                else:
                    cmd.extend(cwl_object.baseCommand)

                cwl_object.baseCommand = cmd
                del cwl_object.hints[index]
                break
        return cwl_object

    def _evaluate_step_inputs(
        self, step: WorkflowStep, inputs: dict[str, Any]
    ) -> dict[str, Any]:
        """Evaluate inputs in Workflow step.

        Parameters
        ----------
            step : WorkflowStep
                Current WorkflowStep
            inputs : dict[str, Any]
                Workflow inputs.

        Returns
        -------
            wf_inputs : dict[str, Any]
                Updated workflow inputs with evaluated step inputs.
        """

        def evaluate_input(
            inp: WorkflowStepInput, input_context: dict, scatter_inputs: dict
        ) -> list | dict[str, Any] | None:
            """Evaluate input value from JS expression or sources.

            Parameters
            ----------
            inp : WorkflowStepInput
                Input 'in_' object
            input_context : dict
                CWL input context
            scatter_inputs : dict
                Scattered inputs

            Returns
            -------
            list | dict[str, Any] | None
                Evaluated CWL File or list of CWL File
            """
            transformed_inputs = deepcopy(self.transformed_inputs)
            inp_id = inp.id.rpartition("#")[2].split("/")[-1]
            if inp.valueFrom:
                # evaluate if it is a JS exp
                if inp.valueFrom.startswith("$("):
                    # handle the case where the input is scattered
                    # First find if the expression contains the scattered input
                    scatter_match = [
                        scat_inp
                        for scat_inp in scatter_inputs.keys()
                        if scat_inp in inp.valueFrom
                    ]
                    if scatter_match:
                        eval_files = []
                        scatter_input_context = deepcopy(input_context)
                        for match in scatter_match:
                            for input in scatter_inputs[match]:
                                # Update context with only one item from the input
                                scatter_input_context[match] = input
                                eval_exp = do_eval(
                                    inp.valueFrom,
                                    scatter_input_context,
                                    outdir=None,
                                    requirements=[JS_REQ],
                                    tmpdir=None,
                                    resources={},
                                )
                                eval_files.append(
                                    format_input_by_type(
                                        eval_exp, get_step_input_type(step, inp_id)
                                    )
                                )
                        return eval_files
                    else:
                        eval_exp = do_eval(
                            inp.valueFrom,
                            input_context,
                            outdir=None,
                            requirements=[JS_REQ],
                            tmpdir=None,
                            resources={},
                        )
                        return format_input_by_type(
                            eval_exp, get_step_input_type(step, inp_id)
                        )
                else:
                    return format_input_by_type(
                        inp.valueFrom, get_step_input_type(step, inp_id)
                    )

            if inp.source:
                # if the input refers to a source from another step output
                # we need to map it using stored interpreted outputs from
                # previous steps
                sources = []
                if isinstance(inp.source, str):
                    sources = [inp.source]
                elif isinstance(inp.source, list):
                    sources = inp.source
                # TODO: do we need to handle the scatter case here?
                for src in sources:
                    inp_source = src.rpartition("#")[2].split("/")
                    if len(inp_source) > 1:
                        step_ref = inp_source[0]
                        source = inp_source[-1]
                        if (
                            source
                            in self.evaluated_steps_outputs.get(step_ref, {}).keys()
                        ):
                            # TODO: can that be a list?
                            return self.evaluated_steps_outputs[step_ref][source]
                    elif inp_source[0] in save(transformed_inputs):
                        return save(transformed_inputs)[inp_source[0]]
                    elif inp_source[0] in input_context:
                        return input_context[inp_source[0]]

        updated_inputs = save(deepcopy(inputs))
        scatter_inputs = self._get_scattered_inputs(step, updated_inputs)
        for inp in step.in_:
            inp_id = inp.id.rpartition("#")[2].split("/")[-1]
            eval_inp = evaluate_input(inp, updated_inputs, scatter_inputs)
            updated_inputs[inp_id] = set_basename_value(eval_inp)

        return updated_inputs

    def _get_scattered_inputs(self, step: WorkflowStep, inputs: dict) -> dict:
        """Retrieve the input names which are scattered.

        Parameters
        ----------
        step : WorkflowStep
            CWL Workflow step object
        inputs : dict
            CWL Input object

        Returns
        -------
        dict:
            scatter input : input value
        """
        scatter_inputs = {}
        if step.scatter:
            if isinstance(step.scatter, list):
                scattered_ids = [
                    scatter_input.rpartition("#")[2].split("/")[-1]
                    for scatter_input in step.scatter
                ]
            else:
                scattered_ids = [step.scatter.rpartition("#")[2].split("/")[-1]]

            for scat_id in scattered_ids:
                # match the scatter input with the inputs
                match = self._find_input_source(step, scat_id)
                scatter_inputs[scat_id] = inputs.get(match, None)
        return scatter_inputs

    def _find_input_source(self, step: WorkflowStep, input: str) -> str | None:
        """Find the source in the step for a given input.

        Parameters
        ----------
            step : WorkflowStep
                The step where inputs are defined.
            input :str
                The name of the input to match.

        Returns
        -------
            (str|None): The source of the input in the step, or None if not found.
        """
        for inp in step.in_:
            # Extract the input ID and check if it matches the scattered input
            inp_id = inp.id.rpartition("#")[2].split("/")[-1]
            if inp_id == input:
                return inp.source.rpartition("#")[2].split("/")[-1]
        return None

    def _evaluate_step_outputs(self, step: WorkflowStep, inputs: dict):
        """Evaluate Workflow Step outputs.

        Parameters
        ----------
            step : WorkflowStep
                WorkflowStep in which evaluate the outputs
            inputs : dict
                Updated inputs with evaluated input from previous steps
        """
        for output in step.run.outputs:
            output_name = output.id.rpartition("#")[2].split("/")[-1]
            glob = output.outputBinding.glob
            if glob:
                if isinstance(glob, str):
                    # the glob is an expression we evaluate it
                    glob_exp = glob if glob.startswith("$(") else None
                    if glob_exp:
                        eval_glob = do_eval(
                            glob,
                            inputs,
                            outdir=None,
                            requirements=[],
                            tmpdir=None,
                            resources={},
                        )
                        if step.scatter and isinstance(eval_glob, list):
                            for out in eval_glob:
                                self.evaluated_steps_outputs[
                                    self.current_step_name
                                ].setdefault(output_name, []).append(
                                    format_output_by_type(out, output.type_)
                                )
                        else:
                            self.evaluated_steps_outputs[self.current_step_name][
                                output_name
                            ] = format_output_by_type(eval_glob, output.type_)
                    else:
                        self.evaluated_steps_outputs[self.current_step_name][
                            output_name
                        ] = format_output_by_type(glob, output.type_)
                elif isinstance(glob, list):
                    self.evaluated_steps_outputs[self.current_step_name].setdefault(
                        output_name, []
                    )
                    for g in glob:
                        if isinstance(g, str):
                            self.evaluated_steps_outputs[self.current_step_name][
                                output_name
                            ].append(format_output_by_type(g, output.type_))

    def _evaluate_expression_tool_outputs(
        self,
        unpacked_cwl_obj: Workflow,
        wf_step: WorkflowStep,
        cwl_inputs: dict,
        step_name: str,
    ):
        """Evaluate the ExpressionTool outputs.

        Parameters
        ----------
            unpacked_cwl_obj : Workflow
                The unpacked CWL workflow.
            wf_step : WorkflowStep
                The Workflow step.
            cwl_inputs : dict
                The CWL inputs dict.
            step_name :str
                The current step name.

        Returns
        -------
            dict:
                A dictionary containing the evaluated ExpressionTool outputs.
        """
        for st in unpacked_cwl_obj.steps:
            if st.id.rpartition("#")[2].split("/")[0] == step_name:
                exp_tool_step = st

        expr_inputs = []
        if wf_step.scatter:
            # In the scattering case we need to verfify the input name
            expr_inputs = self._create_scatter_exptool_inputs(
                wf_step, cwl_inputs, exp_tool_step
            )
        else:
            # In the case where there is no scattering, we assume there is only one single input
            input_name = wf_step.run.inputs[0].id.rpartition("#")[2].split("/")[-1]
            source = exp_tool_step.in_[0].source.rpartition("#")[2].split("/")[-1]
            for inp, value in cwl_inputs.items():
                # Then we match the input name and the input value name
                if inp == source:
                    expr_inputs.append({input_name: value})

        # Finally we evaluate the JS expression
        exptool_outputs = {}
        for inp in expr_inputs:
            inp = fill_defaults(wf_step.run, inp)

            # Evaluate the JS expression
            eval_exp = do_eval(
                wf_step.run.expression,
                save(inp),
                outdir=None,
                requirements=[JS_REQ],
                tmpdir=None,
                resources={},
            )
            for key, val in eval_exp.items():
                output_type = get_step_output_type(wf_step, key)
                if wf_step.scatter:
                    exptool_outputs.setdefault(key, []).append(
                        format_output_by_type(val, output_type)
                    )
                else:
                    exptool_outputs[key] = format_output_by_type(val, output_type)
        return exptool_outputs

    @staticmethod
    def _create_scatter_exptool_inputs(
        wf_step: WorkflowStep, inputs: dict, exp_tool_step: WorkflowStep
    ) -> list[dict]:
        """Create a list of inputs by matching the scatter input names
        and the names in the cwl inputs.

        Parameters
        ----------
            wf_step : WorkflowStep
                The CWL Workflow step.
            inputs: dict
                The CWL inputs.
            exp_tool_step : WorkflowStep
                The Expression Tool step from the unpacked CWL.

        Returns
        -------
            list[dict]:
                Expression Tool inputs list.
        """

        def find_source_in_exp_tool_step(
            exp_tool_step: WorkflowStep, scattered_inp: str
        ) -> str | None:
            """Find the source in the ExpressionTool step for a given scattered input.

            Parameters
            ----------
                exp_tool_step : WorkflowStep
                    The ExpressionTool step where inputs are defined.
                scattered_inp : str
                    The name of the scattered input to match.

            Returns
            -------
                (str|None): The source of the input in the ExpressionTool step, or None if not found.
            """
            for inp in exp_tool_step.in_:
                # Extract the input ID and check if it matches the scattered input
                inp_id = inp.id.rpartition("#")[2].split("/")[-1]
                if inp_id == scattered_inp:
                    return inp.source.rpartition("#")[2].split("/")[-1]
            return None

        expr_inputs = []
        scattered_inp = wf_step.scatter.rpartition("#")[2].split("/")[-1]

        # Find the source corresponding to the scattered input
        source = find_source_in_exp_tool_step(exp_tool_step, scattered_inp)

        # Match the scatter input name with the input value name
        for inp, value in inputs.items():
            if inp == source:
                value = value if isinstance(value, list) else [value]
                expr_inputs.extend({scattered_inp: val} for val in value)

        return expr_inputs

    def _extract_input_files(self, cwl_inputs: dict):
        """Extract input sandboxes/data inputs.

        Parameters
        ----------
        inputs : dict
            CWL inputs
        """
        inputs = deepcopy(cwl_inputs)
        for inp_name, inp_value in inputs.items():
            if isinstance(inp_value, list):
                for index, val in enumerate(inp_value):
                    if isinstance(val, File):
                        self._fill_dirac_inputs(val)
                        # Update the transformed cwl input file
                        self.transformed_inputs[inp_name][index].path = Path(
                            val.path
                        ).name
            elif isinstance(inp_value, File):
                self._fill_dirac_inputs(inp_value)
                # Update the transformed cwl input file
                self.transformed_inputs[inp_name].path = Path(inp_value.path).name

    def _fill_dirac_inputs(self, input_file: File):
        """Fill DIRAC input sanbox and data lists.

        Parameters
        ----------
        input_value : File
            File input
        """
        if input_file.path.startswith(LFN_PREFIX):
            self.input_data.append(
                input_file.path.replace(LFN_PREFIX, LFN_DIRAC_PREFIX)
            )
        else:
            sandbox = Path(input_file.path.replace(LOCAL_PREFIX, ""))
            self.input_sandbox.append(str(sandbox.absolute()))

    def _translate_sandboxes_and_lfns(self, file: File | str) -> tuple[str, bool]:
        """Extract local files as sandboxes and lfns as input data.

        Parameters
        ----------
            file : File | str
                Local file.

        Returns
        -------
            (str, bool):
                A tuple containing a filename and a boolean if it's a lfn (True) or not (False).
        """
        filename = file.path if isinstance(file, File) else file
        if not filename:
            raise KeyError("File path is not defined.")

        is_lfn = filename.startswith(LFN_PREFIX)
        if is_lfn:
            filename = filename.replace(LFN_PREFIX, LFN_DIRAC_PREFIX)
        filename = filename.removeprefix(LOCAL_PREFIX)
        return filename, is_lfn

    def _extract_workflow_outputs(self):
        """Extract the DIRAC outputs (output sandbox/data) from Workflow outputs."""
        for output in self.transformed_cwl.outputs:
            output_sources = output.outputSource
            if isinstance(output_sources, str):
                output_sources = [output_sources]
            if isinstance(output_sources, list):
                for output_src in output_sources:
                    output_source_step = output_src.rpartition("#")[2].split("/")[0]
                    output_source_name = output_src.rpartition("#")[2].split("/")[-1]
                    output_value = self.evaluated_steps_outputs[output_source_step].get(
                        output_source_name, None
                    )

                    if isinstance(output_value, list):
                        for out in output_value:
                            self._fill_dirac_outputs(out)
                    elif output_value:
                        self._fill_dirac_outputs(output_value)

    def _fill_dirac_outputs(self, output_value: dict):
        """Fill DIRAC output sandbox and data lists.

        Parameters
        ----------
            output_value : dict
        """
        if isinstance(output_value, dict) and output_value.get("class", None) == "File":
            if output_value.get("path", "").startswith(LFN_PREFIX):
                self.output_data.append(
                    output_value.get("path", "").replace(LFN_PREFIX, LFN_DIRAC_PREFIX)
                )
            else:
                self.output_sandbox.append(output_value.get("path", ""))

    def _extract_output_files(
        self,
        cwl_obj: CommandLineTool,
        cwl_inputs: dict,
        outputs_to_record: [list | None] = None,
        update_inputs: bool = False,
        always_resolve_output: bool = False,
    ) -> dict:
        """Translate output files into a DIRAC compliant usage.

        Extract local outputs and lfns.
        Remove outputs path prefix.

        Parameters
        ----------
            cwl_obj : CommandLineTool
                The CWL definition.
            cwl_inputs : dict
                CWL inputs.
            outputs_to_record : list|None
                A list of outputs to record.
            update_inputs : bool
                If True, update cwl inputs with output expression (Needed for interpreting JS requirements).
            always_resolve_output : bool
                If True, resolve output expression even if not in the outputs list (Needed for interpreting JS requirements).

        Returns
        -------
            dict:
                Inputs or updated inputs
        """
        inputs = fill_defaults(cwl_obj, cwl_inputs)

        for output in cwl_obj.outputs:
            if not verify_cwl_output_type(output.type_):
                continue

            output_id = (
                output.id.rpartition("#")[2].split("/")[-1] if output.id else None
            )
            if not output_id:
                continue

            glob_list = self._create_glob_list(output)
            for glob in glob_list:
                resolved_glob = self._process_glob(
                    inputs, glob, outputs_to_record, output_id, always_resolve_output
                )
                if update_inputs and output_id not in inputs:
                    inputs = self._update_input_with_output(
                        inputs, resolved_glob, output, output_id
                    )
        return inputs

    @staticmethod
    def _create_glob_list(output: CommandOutputParameter) -> list:
        """
        Create a list of glob expressions.

        Parameters
        ----------
            output : CommandOutputParameter
                The output to process.

        Returns
        -------
            list:
                A list of glob expressions.
        """
        if not output.outputBinding:
            return []
        glob_expr = output.outputBinding.glob
        return [glob_expr] if isinstance(glob_expr, str) else glob_expr or []

    def _process_glob(
        self,
        inputs: dict,
        glob: str,
        outputs_to_record: list | None,
        output_id: str,
        always_resolve_output: bool,
    ) -> str:
        """Evaluate a glob expression and record it if needed.

        Parameters
        ----------
            inputs : dict
                CWL inputs (with added defaults).
            glob : str
                The glob expression.
            outputs_to_record :list
                A list of outputs to record.
            output_id : str
                The current output id.
            always_resolve_output : bool
                If True, resolve output expression even if not in the outputs list (Needed for interpreting JS requirements).
        """
        should_record_output = (
            outputs_to_record is None or output_id in outputs_to_record
        )
        should_eval = (
            should_record_output or always_resolve_output
        ) and glob.startswith("$")
        if should_eval:
            glob = do_eval(
                glob,
                inputs,
                outdir=None,
                requirements=[],
                tmpdir=None,
                resources={},
            )
        if self.current_step_name:
            self.evaluated_steps_outputs[self.current_step_name][output_id] = glob
        if should_record_output:
            if glob.startswith(LFN_PREFIX):
                self.output_data.append(glob.replace(LFN_PREFIX, LFN_DIRAC_PREFIX))
            else:
                self.output_sandbox.append(glob)

        return glob

    @staticmethod
    def _update_input_with_output(
        inputs: dict, resolved_glob: str, output: CommandOutputParameter, output_id: str
    ) -> dict:
        """Inject resolved output value into inputs for JS/parameter expressions.

        Parameters
        ----------
            inputs : dict
                CWL inputs (with added defaults).
            resolved_glob: str
                The resolved glob expression.
            output : CommandOutputParameter
                The output to process.
            output_id : str
                The current output id.

        Returns
        -------
            dict:
                Updated inputs.
        """
        if output.type_ == "File":
            inputs[output_id] = {
                "class": "File",
                "path": resolved_glob,
                "basename": resolved_glob,
            }
        else:
            inputs[output_id] = resolved_glob
        return inputs



from pathlib import Path

import pytest
from cwl_utils.parser.cwl_v1_2 import (
    CommandLineTool,
    CommandOutputArraySchema,
    CommandOutputBinding,
    CommandOutputParameter,
    DockerRequirement,
    File,
    Workflow,
    WorkflowStep,
    StepInputExpressionRequirement,
    WorkflowStepInput,
    WorkflowOutputParameter,
)
from cwl_utils.parser import save

from CTADIRAC.Interfaces.Utilities.CWLTranslator import CWLTranslator
from CTADIRAC.Interfaces.Utilities.CWLUtilities import (
    LFN_PREFIX,
    LFN_DIRAC_PREFIX,
    LOCAL_PREFIX,
)

CVMFS_BASE_PATH = Path("/cvmfs/ctao.dpps.test")
DOCKER_PYTHON_TAG = "harbor/python:tag"


def test_init_class(mocker):
    cwl_workflow = "tests/resources/cwl/single_command_line_tool/process_dl0_dl1.cwl"
    cwl_inputs = (
        "tests/resources/cwl/single_command_line_tool/inputs_process_dl0_dl1.yaml"
    )

    mocker_load_document = mocker.patch(
        "CTADIRAC.Interfaces.Utilities.CWLTranslator.load_document"
    )
    mocker_load_inputfile = mocker.patch(
        "CTADIRAC.Interfaces.Utilities.CWLTranslator.load_inputfile"
    )

    CWLTranslator(cwl_workflow, cwl_inputs)
    mocker_load_document.assert_called_once()
    mocker_load_inputfile.assert_called_once()


def test_translate(mocker):
    # Check if translate method calls the right methods (CLT or Workflow)
    cwl_translator = CWLTranslator.__new__(CWLTranslator)

    # --------------------------------------------
    # Test CommandLineTool
    cwl_translator.transformed_cwl = CommandLineTool.__new__(CommandLineTool)
    mock_translate_clt = mocker.patch(
        "CTADIRAC.Interfaces.Utilities.CWLTranslator.CWLTranslator._translate_clt"
    )

    cwl_translator.translate(CVMFS_BASE_PATH, [])
    mock_translate_clt.assert_called_once_with(CVMFS_BASE_PATH, [])

    # --------------------------------------------
    # Test Workflow
    cwl_translator.transformed_cwl = Workflow.__new__(Workflow)
    mock_translate_workflow = mocker.patch(
        "CTADIRAC.Interfaces.Utilities.CWLTranslator.CWLTranslator._translate_workflow"
    )

    cwl_translator.translate(CVMFS_BASE_PATH, [])
    mock_translate_workflow.assert_called_once_with(CVMFS_BASE_PATH, [])


def test_translate_clt(mocker):
    cwl_translator = CWLTranslator.__new__(CWLTranslator)

    cwl_obj = CommandLineTool(
        inputs=[],
        outputs=[],
        hints=[DockerRequirement(dockerPull=DOCKER_PYTHON_TAG)],
        baseCommand="python",
    )
    cwl_translator.original_inputs = {}
    cwl_translator.transformed_cwl = cwl_obj
    cwl_translator.transformed_inputs = {}
    cwl_translator.output_data = []
    cwl_translator.output_sandbox = []
    cwl_translator.current_step_name = None
    cwl_translator.steps_outputs = {}
    mock_extract_output_files = mocker.patch(
        "CTADIRAC.Interfaces.Utilities.CWLTranslator.CWLTranslator._extract_output_files"
    )

    cwl_translator._translate_clt(CVMFS_BASE_PATH, [])
    assert cwl_translator.transformed_cwl == cwl_obj
    assert cwl_translator.transformed_inputs == {}
    assert cwl_translator.output_data == []
    assert cwl_translator.output_sandbox == []
    mock_extract_output_files.assert_called_once_with(
        cwl_translator.transformed_cwl, cwl_translator.transformed_inputs
    )


def test_translate_workflow():
    def create_workflow(requirement=False):
        return Workflow(
            steps=[
                WorkflowStep(
                    id="/some/path#step_1",
                    in_=[],
                    out=[],
                    run=CommandLineTool(
                        inputs=[],
                        outputs=[],
                        baseCommand="echo CLT1",
                        hints=[DockerRequirement(dockerPull=DOCKER_PYTHON_TAG)],
                    ),
                ),
                WorkflowStep(
                    id="/some/path#step_2",
                    in_=[WorkflowStepInput(id="/some/path#step_4", valueFrom="value")]
                    if requirement
                    else [],
                    out=[],
                    run=CommandLineTool(
                        inputs=[],
                        outputs=[],
                        baseCommand=["echo CLT2"],
                        hints=[DockerRequirement(dockerPull=DOCKER_PYTHON_TAG)],
                    ),
                ),
                WorkflowStep(
                    id="/some/path#step_3",
                    in_=[WorkflowStepInput(id="/some/path#step_5", valueFrom="value")]
                    if requirement
                    else [],
                    out=[],
                    run=CommandLineTool(
                        inputs=[], outputs=[], baseCommand=["echo CLT3"]
                    ),
                ),
            ],
            inputs=[],
            outputs=[],
            requirements=[StepInputExpressionRequirement()] if requirement else [],
        )

    def test_workflow(cwl_workflow):
        cwl_translator.transformed_cwl = cwl_workflow

        cwl_translator._translate_workflow(CVMFS_BASE_PATH, [])

        assert cwl_translator.transformed_cwl == cwl_workflow
        assert cwl_translator.transformed_inputs == {}
        assert cwl_translator.output_data == []
        assert cwl_translator.output_sandbox == []

    cwl_translator = CWLTranslator.__new__(CWLTranslator)

    cwl_translator.original_inputs = {}
    cwl_translator.transformed_inputs = {}
    cwl_translator.output_data = []
    cwl_translator.output_sandbox = []
    cwl_translator.current_step_name = None
    cwl_translator.steps_outputs = {}

    # No requirements workflow
    cwl_workflow = create_workflow()

    # Requirements workflow
    cwl_workflow_req = create_workflow(True)

    test_workflow(cwl_workflow)
    test_workflow(cwl_workflow_req)


@pytest.mark.parametrize(
    ("hints", "base_command", "expected_hints", "expected_base_command"),
    [
        (
            [DockerRequirement(dockerPull=DOCKER_PYTHON_TAG)],
            "python",
            [],
            ["apptainer", "run", str(CVMFS_BASE_PATH / DOCKER_PYTHON_TAG), "python"],
        )
    ],
)
def test_translate_docker_hints(
    hints, base_command, expected_hints, expected_base_command
):
    cwl_translator = CWLTranslator.__new__(CWLTranslator)
    cwl_obj = CommandLineTool(
        inputs=None, outputs=None, hints=hints, baseCommand=base_command
    )

    result = cwl_translator._translate_docker_hints(cwl_obj, CVMFS_BASE_PATH, [])

    assert result.hints == expected_hints
    assert result.baseCommand == expected_base_command


@pytest.mark.parametrize(
    ("input_cwl", "expected_results"),
    [
        (
            {"input1": File(path=LFN_PREFIX + "/ctao/test_lfn_file.txt")},
            {
                "transformed_inputs": {"input1": File(path="test_lfn_file.txt")},
                "input_sandbox": [],
                "input_data": [LFN_DIRAC_PREFIX + "/ctao/test_lfn_file.txt"],
            },
        ),
        (
            {"input1": File(path=LOCAL_PREFIX + "test_local_file.txt")},
            {
                "transformed_inputs": {"input1": File(path="test_local_file.txt")},
                "input_sandbox": [str(Path("test_local_file.txt").absolute())],
                "input_data": [],
            },
        ),
        (
            {
                "input1": [
                    File(path=LFN_PREFIX + "/ctao/test_lfn_file1.txt"),
                    File(path=LOCAL_PREFIX + "test_local_file1.txt"),
                ]
            },
            {
                "transformed_inputs": {
                    "input1": [
                        File(path="test_lfn_file1.txt"),
                        File(path="test_local_file1.txt"),
                    ]
                },
                "input_sandbox": [str(Path("test_local_file1.txt").absolute())],
                "input_data": [LFN_DIRAC_PREFIX + "/ctao/test_lfn_file1.txt"],
            },
        ),
        (
            {
                "input1": File(path=LFN_PREFIX + "/ctao/test_lfn_file2.txt"),
                "input2": File(path=LOCAL_PREFIX + "test_local_file2.txt"),
                "input3": [
                    File(path=LFN_PREFIX + "/ctao/test_lfn_file3.txt"),
                    File(path=LOCAL_PREFIX + "test_local_file3.txt"),
                ],
            },
            {
                "transformed_inputs": {
                    "input1": File(path="test_lfn_file2.txt"),
                    "input2": File(path="test_local_file2.txt"),
                    "input3": [
                        File(path="test_lfn_file3.txt"),
                        File(path="test_local_file3.txt"),
                    ],
                },
                "input_sandbox": [
                    str(Path("test_local_file2.txt").absolute()),
                    str(Path("test_local_file3.txt").absolute()),
                ],
                "input_data": [
                    LFN_DIRAC_PREFIX + "/ctao/test_lfn_file2.txt",
                    LFN_DIRAC_PREFIX + "/ctao/test_lfn_file3.txt",
                ],
            },
        ),
        (
            {
                "input1": [
                    File(path="some/path/test_local_file1.txt"),
                ]
            },
            {
                "transformed_inputs": {
                    "input1": [
                        File(path="test_local_file1.txt"),
                    ]
                },
                "input_sandbox": [
                    str(Path("some/path/test_local_file1.txt").absolute())
                ],
                "input_data": [],
            },
        ),
        (
            {
                "input1": File(path="some/path/test_local_file1.txt"),
            },
            {
                "transformed_inputs": {
                    "input1": File(path="test_local_file1.txt"),
                },
                "input_sandbox": [
                    str(Path("some/path/test_local_file1.txt").absolute())
                ],
                "input_data": [],
            },
        ),
        # Test that a input string containing "lfn" is correctly treated.
        (
            {
                "input1": File(path="some/path/test_local_file1.txt"),
                "input2": f"{LFN_PREFIX}/ctao/path/test.h5",
            },
            {
                "transformed_inputs": {
                    "input1": File(path="test_local_file1.txt"),
                    "input2": f"{LFN_PREFIX}/ctao/path/test.h5",
                },
                "input_sandbox": [
                    str(Path("some/path/test_local_file1.txt").absolute())
                ],
                "input_data": [],
            },
        ),
    ],
)
def test_extract_input_files(input_cwl, expected_results):
    cwl_translator = CWLTranslator.__new__(CWLTranslator)

    cwl_translator.input_data = []
    cwl_translator.input_sandbox = []
    cwl_translator.transformed_inputs = input_cwl

    cwl_translator._extract_input_files(input_cwl)

    assert save(cwl_translator.transformed_inputs) == save(
        expected_results["transformed_inputs"]
    )
    assert cwl_translator.input_sandbox == expected_results["input_sandbox"]
    assert cwl_translator.input_data == expected_results["input_data"]


@pytest.mark.parametrize(
    ("file_input", "expected_result", "expected_lfn"),
    [
        (
            File(path=LFN_PREFIX + "/ctao/test_lfn_file.txt"),
            LFN_DIRAC_PREFIX + "/ctao/test_lfn_file.txt",
            True,
        ),
        (
            File(path=LOCAL_PREFIX + "/home/user/test_local_file.txt"),
            "/home/user/test_local_file.txt",
            False,
        ),
        (
            LFN_PREFIX + "/ctao/test_lfn_str.txt",
            LFN_DIRAC_PREFIX + "/ctao/test_lfn_str.txt",
            True,
        ),
        (
            LOCAL_PREFIX + "/home/user/test_local_str.txt",
            "/home/user/test_local_str.txt",
            False,
        ),
        (File(), None, False),  # This will raise an exception
    ],
)
def test_translate_sandboxes_and_lfns(file_input, expected_result, expected_lfn):
    cwl_translator = CWLTranslator.__new__(CWLTranslator)

    if expected_result is None:
        with pytest.raises(KeyError, match="File path is not defined."):
            cwl_translator._translate_sandboxes_and_lfns(file_input)
    else:
        result, is_lfn = cwl_translator._translate_sandboxes_and_lfns(file_input)
        assert result == expected_result
        assert is_lfn == expected_lfn


@pytest.mark.parametrize(
    ("outputs", "expected_output_sandbox", "expected_output_data"),
    [
        (
            [
                CommandOutputParameter(
                    type_="File",
                    outputBinding=CommandOutputBinding(glob="/path/to/output1.txt"),
                )
            ],
            ["/path/to/output1.txt"],
            [],
        ),
        (
            [
                CommandOutputParameter(
                    type_="File",
                    outputBinding=CommandOutputBinding(
                        glob=LFN_PREFIX + "/path/to/output1.txt"
                    ),
                )
            ],
            [],
            [LFN_DIRAC_PREFIX + "/path/to/output1.txt"],
        ),
        (
            [
                CommandOutputParameter(
                    type_=CommandOutputArraySchema(type_="array", items=File),
                    outputBinding=CommandOutputBinding(
                        glob=[
                            LFN_PREFIX + "/path/to/output1.txt",
                            "/path/to/output2.txt",
                        ]
                    ),
                )
            ],
            ["/path/to/output2.txt"],
            [LFN_DIRAC_PREFIX + "/path/to/output1.txt"],
        ),
    ],
)
def test_extract_output_files(outputs, expected_output_sandbox, expected_output_data):
    cwl_translator = CWLTranslator.__new__(CWLTranslator)

    cwl_obj = CommandLineTool(inputs={}, outputs=outputs)
    cwl_translator.output_sandbox = []
    cwl_translator.output_data = []
    cwl_translator.current_step_name = None
    cwl_translator.steps_outputs = {}

    _ = cwl_translator._extract_output_files(cwl_obj, {})

    assert cwl_translator.output_sandbox == expected_output_sandbox
    assert cwl_translator.output_data == expected_output_data


STEP_INPUT_NAME = "input1"
INPUT_NAME_IN_INPUTS = "input_entry"


@pytest.mark.parametrize(
    ("step", "inputs", "expected_result"),
    [
        (
            WorkflowStep(
                id="/some/path#step_1",
                in_=[
                    WorkflowStepInput(
                        id=f"file:///path/to/#step_1/{STEP_INPUT_NAME}",
                        source=f"file:///path/to/#{INPUT_NAME_IN_INPUTS}",
                    )
                ],
                out=[],
                scatter=[STEP_INPUT_NAME],
                run=CommandLineTool(
                    inputs=[],
                    outputs=[],
                    baseCommand="echo CLT1",
                ),
            ),
            {
                INPUT_NAME_IN_INPUTS: {"class": "File", "path": "/path/to/input"},
                "other_input": "value",
            },
            {STEP_INPUT_NAME: {"class": "File", "path": "/path/to/input"}},
        )
    ],
)
def test_get_scattered_inputs(step, inputs, expected_result):
    cwl_translator = CWLTranslator.__new__(CWLTranslator)
    result = cwl_translator._get_scattered_inputs(step, inputs)
    assert result == expected_result


@pytest.mark.parametrize(
    ("step", "inputs", "expected_result"),
    [
        (
            WorkflowStep(
                id="/some/path#step_1",
                in_=[],
                out=[],
                run=CommandLineTool(
                    inputs=["input1"],
                    outputs=[
                        CommandOutputParameter(
                            id="/some/path#step_1/output1",
                            type_="File",
                            outputBinding=CommandOutputBinding(glob="$(inputs.input1)"),
                        ),
                        CommandOutputParameter(
                            id="/some/path#step_1/output2",
                            type_="string",
                            outputBinding=CommandOutputBinding(glob="static_glob"),
                        ),
                    ],
                    baseCommand="echo CLT1",
                ),
                scatter="input1",
            ),
            {
                "input1": [
                    {"class": "File", "path": "file1.txt"},
                    {"class": "File", "path": "file2.txt"},
                ],
            },
            {
                "step_1": {
                    "output1": [
                        {"class": "File", "path": "file1.txt"},
                        {"class": "File", "path": "file2.txt"},
                    ],
                    "output2": "static_glob",
                }
            },
        ),
        (
            WorkflowStep(
                id="/some/path#step_1",
                in_=[],
                out=[],
                run=CommandLineTool(
                    inputs=["input1"],
                    outputs=[
                        CommandOutputParameter(
                            id="/some/path#step_1/output1",
                            type_="File",
                            outputBinding=CommandOutputBinding(glob="$(inputs.input1)"),
                        ),
                        CommandOutputParameter(
                            id="/some/path#step_1/output2",
                            type_="File[]",
                            outputBinding=CommandOutputBinding(
                                glob=["file3.txt", "file4.txt"]
                            ),
                        ),
                    ],
                    baseCommand="echo CLT1",
                ),
                scatter="input1",
            ),
            {
                "input1": [
                    {"class": "File", "path": "file1.txt"},
                    {"class": "File", "path": "file2.txt"},
                ],
            },
            {
                "step_1": {
                    "output1": [
                        {"class": "File", "path": "file1.txt"},
                        {"class": "File", "path": "file2.txt"},
                    ],
                    "output2": [
                        {"class": "File", "path": "file3.txt"},
                        {"class": "File", "path": "file4.txt"},
                    ],
                }
            },
        ),
    ],
)
def test_evaluate_step_outputs(step, inputs, expected_result):
    cwl_translator = CWLTranslator.__new__(CWLTranslator)
    cwl_translator.current_step_name = "step_1"
    cwl_translator.evaluated_steps_outputs = {}
    cwl_translator.evaluated_steps_outputs.setdefault(
        cwl_translator.current_step_name, {}
    )
    cwl_translator._evaluate_step_outputs(step, inputs)
    assert cwl_translator.evaluated_steps_outputs == expected_result


@pytest.mark.parametrize(
    (
        "transformed_cwl",
        "evaluated_steps_outputs",
        "expected_output_sandbox",
        "expected_output_data",
    ),
    [
        (
            Workflow(
                inputs=[],
                outputs=[
                    WorkflowOutputParameter(
                        id="/some/path#output1",
                        outputSource="/some/path#step_1/output1",
                        type_="File",
                    ),
                    WorkflowOutputParameter(
                        id="/some/path#output2",
                        outputSource="/some/path#step_2/output2",
                        type_="File[]",
                    ),
                ],
                steps=[
                    WorkflowStep(
                        id="/some/path#step_1",
                        in_=[],
                        out=[],
                        run=CommandLineTool(
                            inputs=[],
                            outputs=[],
                            baseCommand="echo CLT1",
                            hints=[DockerRequirement(dockerPull=DOCKER_PYTHON_TAG)],
                        ),
                    )
                ],
            ),
            {
                "step_1": {
                    "output1": {"class": "File", "path": "/local/path/file1.txt"},
                },
                "step_2": {
                    "output2": [
                        {"class": "File", "path": f"{LFN_PREFIX}remote/path/file2.txt"},
                        {"class": "File", "path": "/local/path/file3.txt"},
                    ],
                },
            },
            ["/local/path/file1.txt", "/local/path/file3.txt"],
            [f"{LFN_DIRAC_PREFIX}remote/path/file2.txt"],
        ),
    ],
)
def test_extract_workflow_outputs(
    transformed_cwl,
    evaluated_steps_outputs,
    expected_output_sandbox,
    expected_output_data,
):
    # Initialize the class under test
    instance = CWLTranslator.__new__(CWLTranslator)

    # Set up the necessary state
    instance.transformed_cwl = transformed_cwl
    instance.evaluated_steps_outputs = evaluated_steps_outputs
    instance.output_sandbox = []
    instance.output_data = []

    # Call the method under test
    instance._extract_workflow_outputs()

    # Assert the results
    assert instance.output_sandbox == expected_output_sandbox
    assert instance.output_data == expected_output_data



from copy import deepcopy
from typing import Union
from pathlib import Path
from cwl_utils.parser import (
    File,
    InputArraySchema,
    OutputArraySchema,
    Workflow,
    WorkflowStep,
    CommandLineTool,
    ExpressionTool,
)
from cwl_utils.parser.cwl_v1_2 import CommandInputArraySchema

LFN_PREFIX = "lfn://"
LFN_DIRAC_PREFIX = "LFN:"
LOCAL_PREFIX = "file://"
JS_REQ = {"class": "InlineJavascriptRequirement"}


def verify_cwl_output_type(
    output_type: Union[str, OutputArraySchema, list[Union[str, OutputArraySchema]]]
) -> bool:
    """Check if output type is a "File" str or OutputArraySchema (or a list containing them).

    Parameters
    ----------
    output_type : Union[str, OutputArraySchema, List[Union[str, OutputArraySchema]]]
        The output type.
    """
    if isinstance(output_type, list):
        return any(t == "File" or isinstance(t, OutputArraySchema) for t in output_type)
    return output_type == "File" or isinstance(output_type, OutputArraySchema)


def get_current_step_obj(cwl_obj: Workflow, step_name: str):
    """Get the current step object by step name.

    Parameters
    ----------
    cwl_obj : Workflow
        The CWL workflow object containing steps.
    step_name : str
        The name of the step to retrieve.

    Returns
    -------
    step : WorkflowStep | None
        The matched step object, or None if not found.
    """
    for step in cwl_obj.steps:
        if step.id.rpartition("#")[2].split("/")[0] == step_name:
            return step
    return None


def fill_defaults(cwl: Union[Workflow, CommandLineTool, ExpressionTool], inputs: dict):
    """Fill defaults from CWL inputs into inputs dict.
    This is needed for evaluating expressions later on.

    Parameters
    ----------
    cwl : Union[Workflow, CommandLineTool, ExpressionTool]
        The CWL definition
    inputs : dict
        User provided inputs.

    Returns
    -------
    inputs : dict
        Inputs with additional values filled from CWL defaults
    """
    updated_inputs = deepcopy(inputs)

    def fill_input_inputs(step):
        for inp in step.inputs:
            key = inp.id.rpartition("#")[2].split("/")[-1]
            if inp.default is not None:  # and key not in updated_inputs
                updated_inputs[key] = inp.default

    if isinstance(cwl, Workflow):
        for step in cwl.steps:
            fill_input_inputs(step.run)
    elif isinstance(cwl, (CommandLineTool, ExpressionTool)):
        fill_input_inputs(cwl)

    return updated_inputs


def set_input_file_basename(inputs: dict) -> None:
    """Ensure input Files have basename set.

    Parameters
    ----------
    inputs : dict
        cwl inputs
    """
    for inp in inputs.values():
        inp = set_basename_value(inp)
    return inputs


def set_basename_value(input: dict):
    """Set basename field of a cwl File.

    Parameters
    ----------
    input : dict | File
        CWL input

    Returns
    -------
    dict | File
        CWL input
    """
    if isinstance(input, File) and not input.basename:
        input.basename = Path(input.path).name
    elif isinstance(input, list):
        for val in input:
            if isinstance(val, File) and not val.basename:
                val.basename = Path(val.path).name
    # ensure it works with dict inputs
    elif isinstance(input, dict):
        if input.get("class", None) == "File" and not input.get("basename", None):
            input["basename"] = Path(input.get("path", "")).name
        if input.get("class", None) == "List" and not input.get("basename", None):
            for val in input:
                if input.get("class", None) == "File" and not input.get(
                    "basename", None
                ):
                    input["basename"] = Path(input.get("path", "")).name
    return input


def get_step_input_type(step: WorkflowStep, input_name: str):
    """Retrieve the step input type.

    Parameters
    ----------
    step : WorkflowStep
    input_name : str

    Returns
    -------
    Any:
        input type
    """
    for input in step.run.inputs:
        if input.id.rpartition("#")[2].split("/")[-1] == input_name:
            return input.type_


def get_step_output_type(step: WorkflowStep, output_name: str):
    """Retrieve the step output type.

    Parameters
    ----------
    step : WorkflowStep
    output_name : str

    Returns
    -------
    Any:
        Output type
    """
    for output in step.run.outputs:
        if output.id.rpartition("#")[2].split("/")[-1] == output_name:
            return output.type_


def format_input_by_type(
    input: str, in_type: list | str
) -> str | dict[str, str] | list[dict[str, str]]:
    """Format the input value by type

    Parameters
    ----------
    input : str
        input name
    in_type : list | str
        _description_

    Returns
    -------
    str | dict[str, str] | list[dict[str, str]]
        formatted input
    """
    if isinstance(in_type, list):
        if "null" in in_type:
            in_type.remove("null")
        if len(in_type) == 1:
            in_type = in_type[0]
        elif "File" in in_type:
            in_type = "File"
            # Is there any other cases than [File, string]
    if isinstance(in_type, str):
        if in_type in ["boolean", "string", "int"]:
            return input
        elif in_type == "File":
            return {"class": "File", "path": input}
    elif isinstance(in_type, (CommandInputArraySchema, InputArraySchema)):
        if in_type.type_ == "File":
            return {"class": "File", "path": input}
        elif in_type.type_ == "string":
            return input
    else:
        return input


def format_output_by_type(
    output: str, out_type: list | str
) -> str | dict[str, str] | None:
    """Format output value by type

    Parameters
    ----------
    output : str
        Output name
    out_type : list | str
        Output type

    Returns
    -------
    str | dict[str, str] | None
        Formatted output
    """
    if isinstance(output, str):
        if isinstance(out_type, str):
            if out_type in ["boolean", "string", "int"]:
                return output
            elif out_type in ["File", "File[]"]:
                return {"class": "File", "path": output}
        elif isinstance(out_type, OutputArraySchema):
            if out_type.type_ == "File":
                return {"class": "File", "path": output}
            elif out_type.type_ == "string":
                return output
        else:
            return output
    else:
        return output



#!/usr/bin/env python

__RCSID__ = "$Id$"

# generic imports
import os
from multiprocessing import Pool

# DIRAC imports
import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script
from DIRAC.DataManagementSystem.Client.DataManager import DataManager

from CTADIRAC.Core.Utilities.tool_box import read_inputs_from_file

Script.setUsageMessage(
    """
Bulk removal of a list of replicas or a single replicas from a given SE
Usage:
   cta-prod-remove-replica [options] <lfn> <SE>
   cta-prod-remove-replica [options] <ascii file with lfn list> <SE>
"""
)

Script.parseCommandLine(ignoreErrors=True)

DIRAC.initialize()

args = Script.getPositionalArgs()
if len(args) > 1:
    infile = args[0]
    SE = args[1]
else:
    Script.showHelp()

if os.path.isfile(infile):
    infileList = read_inputs_from_file(infile)
else:
    infileList = [infile]


def removeReplica(lfn):
    voName = lfn.split("/")[1]
    if voName not in ["ctao", "vo.cta.in2p3.fr"]:
        message = (
            f"Wrong lfn: path must start with vo name (ctao or vo.cta.in2p3.fr):\n{lfn}"
        )
        gLogger.error(message)
        return
    dm = DataManager(vo=voName)
    res = dm.removeReplica(SE, lfn)
    if not res["OK"]:
        message = res["Message"]
        gLogger.error(f"Error removing replica {lfn} from {SE}:\n{message}")
        return


@Script()
def main():
    p = Pool(10)
    p.map(removeReplica, infileList)


if __name__ == "__main__":
    main()



import json
from typing import Any

import DIRAC
from ruamel.yaml import YAML

from CTADIRAC.Core.Utilities.tool_box import DATA_LEVEL_METADATA_ID
from CTADIRAC.Interfaces.API.CTAJob import CTAJob


class SimPipeJob(CTAJob):
    """Generic MCPipe Job class"""

    def __init__(self, we_type: str = "simpipe") -> None:
        super().__init__(we_type=we_type)
        self.setName("MC_Generation")
        self.setType("MCSimulation")
        self.package = "simpipe"
        self.compiler = "gcc114_default"
        self.program_category = "tel_sim"
        self.prog_name = "simpipe"
        self.container = True
        self.run_number_offset = 0
        self.run_number = "@{JOB_ID}"
        self.data_level: int = DATA_LEVEL_METADATA_ID["R1"]
        self.base_path = "/ctao/MC/PROD6/"
        self.sct = False
        self.only_corsika = False
        self.array_layout = ""
        self.zenith_angle = 20.0
        self._azimuth_angle = None
        self.software_category = "simulations"
        self.particle = None
        self.view_cone = None
        self.output_directory = None
        self.simpipe_config_options = None
        self.save_reduced_event_lists = False
        self.output_reduced_event_lists_type = "reduced_event_lists"

    def set_simpipe_config(self, value) -> None:
        config = self._load_config_file(value)
        self._apply_config(config)

    def _load_config_file(self, config_file: str) -> dict:
        yaml = YAML(typ="safe")
        with open(config_file) as file:
            return yaml.load(file)

    def _apply_config(self, config: dict) -> None:
        """Apply configuration settings to the class attributes."""

        # Define a generic handler for setting attributes
        def set_attribute(attr_name, value):
            setattr(self, attr_name, value)

        # Define a mapping of keys to handler methods or attributes
        handlers = {
            "array_layout_name": self._set_array_layout,
            "site": self._set_site,
            "model_version": lambda value: setattr(
                self, "model_version", value if isinstance(value, list) else [value]
            ),
            "simulation_software": self._set_simulation_software,
            "pack_for_grid_register": lambda value: setattr(
                self, "output_directory", value
            ),
            "primary": lambda value: setattr(self, "particle", value.lower()),
            "save_reduced_event_lists": lambda value: setattr(
                self, "save_reduced_event_lists", bool(value)
            ),
            "view_cone": lambda value: setattr(
                self, "view_cone", str(value).replace(" ", "_")
            ),
        }

        # Add generic handlers for attributes that map directly
        direct_mappings = ["zenith_angle", "azimuth_angle"]
        for key in direct_mappings:
            handlers[key] = lambda value, attr=key: set_attribute(attr, value)

        # Get the configuration in a dict to serialize it for the simpipe wrapper
        self.simpipe_config_options = {}
        for key, value in config.items():
            key_lower = key.lower()
            if key_lower in handlers:
                handlers[key_lower](value)
            elif key_lower in {"run_number_offset", "run_number"}:
                DIRAC.gLogger.error(
                    f"The {key_lower} must be set in the DIRAC config file, "
                    "not the SimPipe config file."
                )
                DIRAC.exit(-1)

            # Store the configuration in simpipe_config_options
            self.simpipe_config_options[key_lower] = value

    def _set_array_layout(self, value: str) -> None:
        """Set the array layout and SCT flag."""
        self.array_layout = value
        if "_scts" in value:
            self.sct = True
            # TODO: choose also the right container for this case

    def _set_site(self, value: str) -> None:
        """Set the site based on the configuration."""
        if value == "North":
            self.site = "LaPalma"
        elif value == "South":
            self.site = "Paranal"
        else:
            DIRAC.gLogger.error(f"Unknown site: {value}")
            DIRAC.exit(-1)

    @property
    def azimuth_angle(self) -> float:
        """Get the azimuth angle."""
        if self._azimuth_angle is None:
            DIRAC.gLogger.error(
                "Azimuth angle is not set. "
                "Please set it using the azimuth_angle property."
            )
            DIRAC.exit(-1)
        return self._azimuth_angle

    @azimuth_angle.setter
    def azimuth_angle(self, value: float) -> None:
        try:
            value = float(value)
            self._azimuth_angle = value
        except ValueError:
            azimuth_map = {
                "north": 0,
                "south": 180,
                "east": 90,
                "west": 270,
            }
            azimuth_angle = value.lower()
            if azimuth_angle in azimuth_map:
                self._azimuth_angle = azimuth_map[azimuth_angle]
            else:
                DIRAC.gLogger.error(
                    f"Unknown azimuth angle: {value}. "
                    f"Options are: {list(azimuth_map.keys())}"
                )
                DIRAC.exit(-1)

    def _set_simulation_software(self, value: str) -> None:
        """Set simulation software-specific attributes."""
        if value == "corsika":
            self.only_corsika = True
            self.program_category = "airshower_sim"
            self.prog_name = "corsika"

    def build_file_metadata(
        self, combination, propagate_run_number=True
    ) -> tuple[dict[str, str], Any, str, str]:
        file_meta_data: dict[str, str] = {}
        if propagate_run_number:
            file_meta_data["runNumber"] = self.run_number
            if self.run_number != "@{JOB_ID}":
                file_meta_data["runNumber"] += self.run_number_offset

        for key, value in combination.items():
            file_meta_data[key] = value

        return file_meta_data

    def run_simpipe(self, debug=False) -> None:
        """
        Run CORSIKA/sim_telarray simulation step
        """
        simpipe_config_json = json.dumps(self.simpipe_config_options)
        prod_exe = "dirac_simpipe_simulate_prod_wrapper"
        prod_args = (
            f"{self.run_number_offset} {self.run_number} '{simpipe_config_json}'"
        )

        step = self.setExecutable(
            prod_exe,
            arguments=prod_args,
            logFile="SimPipe_Log.txt",
            modulesList=["cta_script"],
        )
        step["Value"]["name"] = "Step_SimPipe"
        step["Value"]["descr_short"] = "Run SimPipe simulation step"

    def init_debug_step(self) -> None:
        super().init_debug_step()
        step = self.setExecutable("/bin/env", logFile="Env_Log.txt")
        step["Value"]["name"] = "Step_Env"
        step["Value"]["descr_short"] = "Dump environment"

    def upload_and_register_file(
        self,
        meta_data_json,
        file_meta_data_json,
        file_pattern,
        log_str,
        data_type: str = "data",
    ) -> None:
        """
        Upload and register files to SE and register them in DFC.

        Args:
            meta_data_json: JSON string containing metadata
            file_meta_data_json: JSON string containing file metadata
            file_pattern: Pattern for files to upload
            log_str: String identifier for logging
            data_type: Type of data ('data', 'log', 'reduced_event_lists')
        """
        # Define configuration for different data types
        type_config = {
            "data": {
                "output_type_attr": "output_data_type",
                "management_type": "DataManagement",
                "step_name": "Step_DataManagement",
                "description": "Save data files to SE and register them in DFC",
            },
            "log": {
                "output_type_attr": "output_log_type",
                "management_type": "LogManagement",
                "step_name": "Step_LogManagement",
                "description": "Save log files to SE and register them in DFC",
            },
            "reduced_event_lists": {
                "output_type_attr": "output_reduced_event_lists_type",
                "management_type": "ReducedEventListsManagement",
                "step_name": "Step_ReducedEventListsManagement",
                "description": (
                    "Save reduced event lists to SE and register them in DFC"
                ),
            },
        }

        if data_type not in type_config:
            raise ValueError(
                f"Unknown data type: {data_type}. "
                f"Supported types: {list(type_config.keys())}"
            )

        config = type_config[data_type]
        output_data_type = getattr(self, config["output_type_attr"])
        management_type = config["management_type"]
        step_name = config["step_name"]
        description = config["description"]

        step = self.setExecutable(
            "cta-prod-managedata",
            arguments=(
                f"'{meta_data_json}' '{file_meta_data_json}' {self.base_path} "
                f"'{file_pattern}' {self.package} {self.program_category} "
                f"'{self.catalogs}' {output_data_type}"
            ),
            logFile=f"{management_type}_{log_str}_Log.txt",
        )

        step["Value"]["name"] = step_name
        step["Value"]["descr_short"] = description

    def set_metadata_and_register_data(self, propagate_run_number=True) -> None:
        meta_data_json: str = json.dumps(self.output_metadata)

        # Iterate over each model_version entry
        for model_version in self.model_version:
            combination = {"model_version": model_version}

            # Build file metadata for the current model_version
            file_meta_data = self.build_file_metadata(combination, propagate_run_number)
            file_meta_data_json = json.dumps(file_meta_data)

            data_output_pattern = f"{self.output_directory}/*_{model_version}_*.zst"
            if self.only_corsika:
                data_output_pattern = (
                    f"{self.output_directory}/*_{model_version}_*.corsika.zst"
                )
            log_str = f"{model_version}"

            # Upload and register data files for the current model_version
            self.upload_and_register_file(
                meta_data_json,
                file_meta_data_json,
                data_output_pattern,
                log_str,
                data_type="data",
            )

            # Upload reduced event lists files if available
            if self.save_reduced_event_lists:
                reduced_data_output_pattern = (
                    f"{self.output_directory}/*_{model_version}_*"
                    ".reduced_event_data.hdf5"
                )

                # Upload and register reduced event lists files
                self.upload_and_register_file(
                    meta_data_json,
                    file_meta_data_json,
                    reduced_data_output_pattern,
                    f"{log_str}_reduced_event_lists",
                    data_type="reduced_event_lists",
                )

            # Use model_version in log file patterns
            log_file_pattern = (
                f"{self.output_directory}/*_{model_version}_*.log_hist.tar.gz"
            )
            if self.only_corsika:
                log_file_pattern = (
                    f"{self.output_directory}/*_{model_version}_*.corsika.log.gz"
                )

            # Upload and register log files for the current model_version
            self.upload_and_register_file(
                meta_data_json,
                file_meta_data_json,
                log_file_pattern,
                log_str,
                data_type="log",
            )

    def run_dedicated_software(self) -> None:
        self.run_simpipe()

    def set_executable_sequence(
        self, debug: bool = False, define_n_showers=True
    ) -> None:
        super().set_executable_sequence(debug=debug)



import xml.etree.ElementTree as ET
from pathlib import Path

import pytest
from cwl_utils.parser import save
from CTADIRAC.Interfaces.API.CWLJob import CWLJob
from CTADIRAC.Interfaces.Utilities.CWLUtilities import (
    LFN_DIRAC_PREFIX,
    LFN_PREFIX,
)

INPUT_DATA = ["/ctao/user/MC/prod.sim", "/ctao/user/MC/prod2.sim"]
INPUT_SANDBOX = ["/a/local/MC/simulation.py", "/path/to/MC.prod3.sim"]
CWL_INPUTS_EXAMPLE = f"""
local_script:
  - class: File
    path: {INPUT_SANDBOX[0]}
input_as_lfn:
  - class: File
    path: {LFN_PREFIX}{INPUT_DATA[0]}
input_as_lfn_2:
  - class: File
    path: {LFN_PREFIX}{INPUT_DATA[1]}
  - class: File
    path: {INPUT_SANDBOX[1]}
dataset: "dataset://path/to/data"
input_param: "a random param"
"""
OUTPUT_DATA = ["/ctao/user/MC/fit*.out", "/ctao/usr/MC/data.sim"]
OUTPUT_SANDBOX = ["/path/to/test*.out", "/path/to/data_2.sim", "*.txt"]
BASE_COMMAND = "python"
IMAGE = "harbor.cta-observatory.org/proxy_cache/library/python:3.12-slim"
CWL_WORKFLOW_EXAMPLE = f"""
cwlVersion: v1.2
class: CommandLineTool
doc: |
      Test using local input/output data treated as Dirac
      input/output sandboxes.

inputs:
  local_script:
    type: File
    inputBinding:
      position: 1
  input_as_lfn:
    type: File?
    inputBinding:
      position: 2
  input_as_lfn_2:
    type: File[]
    inputBinding:
      position: 3
  dataset:
    type: [File, string]
    inputBinding:
      position: 4
  input_param:
    type: string
    inputBinding:
      position: 5

outputs:
  output_as_sb:
    type: File[]?
    outputBinding:
      glob: ["{OUTPUT_SANDBOX[0]}"]
  output_as_lfn:
    type: File?
    label: "LFN wildcards"
    outputBinding:
      glob: "{LFN_PREFIX}{OUTPUT_DATA[0]}"
  output_as_lfn_2:
    type: File[]
    label: "LFN files list"
    outputBinding:
      glob:
        - {LFN_PREFIX}{OUTPUT_DATA[1]}
        - {OUTPUT_SANDBOX[1]}
  output_as_array:
    type:
      type: array
      items: File
    outputBinding:
      glob: "{OUTPUT_SANDBOX[2]}"

baseCommand: ["{BASE_COMMAND}"]

hints:
  DockerRequirement:
    dockerPull: {IMAGE}
"""

CWL_WORKFLOW_EXAMPLE_NO_INPUT = """
cwlVersion: v1.2
class: CommandLineTool
doc: |
      Test using local input/output data treated as Dirac
      input/output sandboxes.

inputs:
  local_script:
    type: string
    inputBinding:
      position: 1

outputs:
  output_as_sb:
    type: File[]?
    outputBinding:
      glob: ["/path/to/output_1.txt"]

baseCommand: ["base_command"]
"""
CWL_WORKFLOW_EXAMPLE_NO_INPUT_INPUTS = """
local_script: "test"
"""
CVMFS_BASE_PATH = Path("/cvmfs/ctao.dpps.test/")

# Input files
INPUT_1 = "gamma_1.dl1_img.h5"
INPUT_2 = "gamma_2.dl1_img.h5"

# Output files
MERGED_OUTPUT = "merged.dl1.h5"  # merged_output
MERGE_LOG = "ctapipe-merge.log"  # merge_log
MERGE_PROVENANCE_LOG = "ctapipe-merge.provenance.log"  # merge_provenance_log
INTERMEDIATE_LOG = "gamma_1.dl1.h5.log"  # intermediate_log
INTERMEDIATE_PROVENANCE_LOG = "gamma_1.dl1.h5.provlog"  # intermediate_log_provenance
INTERMEDIATE_LOG_2 = "gamma_2.dl1.h5.log"
INTERMEDIATE_PROVENANCE_LOG_2 = "gamma_2.dl1.h5.provlog"


@pytest.fixture
def mock_submit_job(mocker):
    return mocker.patch(
        "CTADIRAC.Interfaces.API.CWLJob.Dirac.submitJob",
        side_effect=lambda self: self._toXML(),
    )


def get_list(xml, name):
    element = xml.find(f".//Parameter[@name='{name}']/value")
    if element is None or element.text is None:
        return []
    return element.text.split(";")


@pytest.mark.parametrize(
    (
        "cwl_worflow",
        "cwl_inputs",
        "expected_input_data",
        "expected_input_sandbox",
        "expected_output_data",
        "expected_output_sandbox",
    ),
    [
        (
            CWL_WORKFLOW_EXAMPLE,
            CWL_INPUTS_EXAMPLE,
            [f"{LFN_DIRAC_PREFIX}{data}" for data in INPUT_DATA],
            INPUT_SANDBOX,
            [f"{LFN_DIRAC_PREFIX}{data}" for data in OUTPUT_DATA],
            OUTPUT_SANDBOX,
        ),
        (
            CWL_WORKFLOW_EXAMPLE_NO_INPUT,
            CWL_WORKFLOW_EXAMPLE_NO_INPUT_INPUTS,
            [],
            [],
            [],
            ["/path/to/output_1.txt"],
        ),
    ],
)
def test_cwl_job_submit(
    mock_submit_job,
    tmp_path,
    cwl_worflow,
    cwl_inputs,
    expected_input_data,
    expected_input_sandbox,
    expected_output_data,
    expected_output_sandbox,
):
    cwl_workflow_example = tmp_path / "cwl_workflow_example.cwl"
    cwl_workflow_example.write_text(cwl_worflow)
    cwl_inputs_example = tmp_path / "cwl_inputs_example.cwl"
    cwl_inputs_example.write_text(cwl_inputs)

    job = CWLJob(
        cwl_workflow=cwl_workflow_example,
        cwl_inputs=cwl_inputs_example,
        cvmfs_base_path=CVMFS_BASE_PATH,
        output_se="TEST_SE",
    )

    result = job.submit()
    mock_submit_job.assert_called_once_with(job)
    result_xml = ET.fromstring(result)

    input_data = get_list(result_xml, "InputData")
    assert input_data == expected_input_data

    output_sandbox = get_list(result_xml, "OutputSandbox")
    assert set(expected_output_sandbox).issubset(output_sandbox)

    input_sandbox = get_list(result_xml, "InputSandbox")
    assert set(expected_input_sandbox).issubset(input_sandbox)

    output_data = get_list(result_xml, "OutputData")
    assert output_data == expected_output_data

    if expected_output_data:
        output_se_parameter = result_xml.find(".//Parameter[@name='OutputSE']")
        assert output_se_parameter is not None
        output_se = output_se_parameter.find("value")
        assert output_se is not None
        assert output_se.text == "TEST_SE"

    if executable := result_xml.find(
        ".//StepInstance/Parameter[@name='executable']/value"
    ):
        executable = executable.text
        assert executable == "cwltool"


def test_datapipe_cwl_job(mock_submit_job):
    job = CWLJob(
        "tests/resources/cwl/single_command_line_tool/process_dl0_dl1.cwl",
        "tests/resources/cwl/single_command_line_tool/inputs_process_dl0_dl1.yaml",
        cvmfs_base_path=CVMFS_BASE_PATH,
    )

    # these are set via inputs
    assert len(job.input_data) == 1
    assert (
        job.input_data[0]
        == "LFN:/ctao/simpipe/prod6/gamma-diffuse/010xxx/gamma_cone10_run010000.simtel.zst"
    )

    assert save(job.transformed_inputs) == {
        "dl0": {"class": "File", "path": "gamma_cone10_run010000.simtel.zst"},
        "processing_config": {"class": "File", "path": "process_config.yaml"},
        "dl1_filename": "lfn:///ctao/datapipe/test.dl1.h5",
    }

    result = job.submit()
    mock_submit_job.assert_called_once_with(job)
    result_xml = ET.fromstring(result)

    input_data = get_list(result_xml, "InputData")
    assert len(input_data) == 1
    assert (
        input_data[0]
        == "LFN:/ctao/simpipe/prod6/gamma-diffuse/010xxx/gamma_cone10_run010000.simtel.zst"
    )

    input_sandbox = get_list(result_xml, "InputSandbox")
    assert len(input_sandbox) == 3
    assert "process_config.yaml" in input_sandbox[-1]

    output_data = get_list(result_xml, "OutputData")
    assert len(output_data) == 1
    assert output_data[0] == "LFN:/ctao/datapipe/test.dl1.h5"

    output_sandbox = get_list(result_xml, "OutputSandbox")
    assert len(output_sandbox) == 2
    assert output_sandbox == [
        "ctapipe-process.log",
        "ctapipe-process_dl0_dl1.provenance.log",
    ]


def test_datapipe_cwl_workflow_job(mock_submit_job):
    job = CWLJob(
        "tests/resources/cwl/scatter_feat_requirement/process_dl0_dl1_multiple.cwl",
        "tests/resources/cwl/scatter_feat_requirement/inputs_dl0_dl1_multiple.yaml",
        cvmfs_base_path=CVMFS_BASE_PATH,
    )
    assert len(job.input_data) == 0

    assert len(job.input_sandbox) == 2
    assert INPUT_1 in job.input_sandbox[0]
    assert INPUT_2 in job.input_sandbox[1]

    assert len(job.output_data) == 0
    assert job.output_sandbox == [
        MERGED_OUTPUT,
        MERGE_LOG,
        MERGE_PROVENANCE_LOG,
    ]

    result = job.submit()
    mock_submit_job.assert_called_once_with(job)
    result_xml = ET.fromstring(result)

    input_data = get_list(result_xml, "InputData")
    assert input_data == []

    input_sandbox = get_list(result_xml, "InputSandbox")
    assert len(input_sandbox) == 4  # cwl + inputs + input_sandbox

    output_data = get_list(result_xml, "OutputData")
    assert len(output_data) == 0

    output_sandbox = get_list(result_xml, "OutputSandbox")
    assert len(output_sandbox) == 3
    assert output_sandbox[0] == MERGED_OUTPUT


# TODO: parametrize with the above test?
def test_datapipe_dl0_dl2_workflow(mock_submit_job):
    """Handling JS requirements."""
    job = CWLJob(
        "tests/resources/cwl/step_input_requirement/workflow_dl0_to_dl2.cwl",
        "tests/resources/cwl/step_input_requirement/inputs_workflow_dl0_to_dl2.yaml",
        cvmfs_base_path=CVMFS_BASE_PATH,
    )
    assert len(job.input_data) == 1
    assert job.input_data == [
        "LFN:/ctao/simpipe/prod6/gamma-diffuse/010xxx/gamma_prod5.simtel.zst"
    ]

    assert len(job.input_sandbox) == 0  # if no processing config

    assert len(job.output_data) == 0

    assert job.output_sandbox == ["gamma_prod5.dl1.h5", "gamma_prod5.dl2.h5"]

    result = job.submit()
    mock_submit_job.assert_called_once_with(job)
    result_xml = ET.fromstring(result)

    input_data = get_list(result_xml, "InputData")
    assert input_data == [
        "LFN:/ctao/simpipe/prod6/gamma-diffuse/010xxx/gamma_prod5.simtel.zst"
    ]

    input_sandbox = get_list(result_xml, "InputSandbox")
    assert len(input_sandbox) == 2  # cwl + inputs

    output_data = get_list(result_xml, "OutputData")
    assert len(output_data) == 0

    output_sandbox = get_list(result_xml, "OutputSandbox")
    assert output_sandbox == ["gamma_prod5.dl1.h5", "gamma_prod5.dl2.h5"]


def test_scattered_expression_tool(mock_submit_job):
    job = CWLJob(
        "tests/resources/cwl/expression_tool_scatter/process_dl0_dl1_multiple.cwl",
        "tests/resources/cwl/expression_tool_scatter/inputs_process_multiple.yaml",
        cvmfs_base_path=CVMFS_BASE_PATH,
    )

    assert len(job.input_data) == 0

    assert len(job.input_sandbox) == 2
    assert INPUT_1 in job.input_sandbox[0]
    assert INPUT_2 in job.input_sandbox[1]

    assert len(job.output_data) == 0
    assert set(job.output_sandbox) == {
        MERGED_OUTPUT,
        MERGE_LOG,
        MERGE_PROVENANCE_LOG,
        INTERMEDIATE_LOG,
        INTERMEDIATE_LOG_2,
        INTERMEDIATE_PROVENANCE_LOG,
        INTERMEDIATE_PROVENANCE_LOG_2,
    }

    result = job.submit()
    mock_submit_job.assert_called_once_with(job)
    result_xml = ET.fromstring(result)

    input_data = get_list(result_xml, "InputData")
    assert input_data == []

    input_sandbox = get_list(result_xml, "InputSandbox")
    assert len(input_sandbox) == 4  # cwl + inputs + input_sandbox

    output_data = get_list(result_xml, "OutputData")
    assert len(output_data) == 0

    output_sandbox = get_list(result_xml, "OutputSandbox")
    assert len(output_sandbox) == 7


def test_expression_tool(mock_submit_job):
    job = CWLJob(
        "tests/resources/cwl/expression_tool/process_dl0_dl1_multiple.cwl",
        "tests/resources/cwl/expression_tool/inputs_process_multiple.yaml",
        cvmfs_base_path=CVMFS_BASE_PATH,
    )

    assert len(job.input_data) == 0

    assert len(job.input_sandbox) == 1
    assert INPUT_1 in job.input_sandbox[0]

    assert len(job.output_data) == 0
    assert set(job.output_sandbox) == {
        MERGED_OUTPUT,
        MERGE_LOG,
        MERGE_PROVENANCE_LOG,
        INTERMEDIATE_LOG,
        INTERMEDIATE_PROVENANCE_LOG,
    }

    result = job.submit()
    mock_submit_job.assert_called_once_with(job)
    result_xml = ET.fromstring(result)

    input_data = get_list(result_xml, "InputData")
    assert input_data == []

    input_sandbox = get_list(result_xml, "InputSandbox")
    assert len(input_sandbox) == 3  # cwl + inputs + input_sandbox

    output_data = get_list(result_xml, "OutputData")
    assert len(output_data) == 0

    output_sandbox = get_list(result_xml, "OutputSandbox")
    assert len(output_sandbox) == 5


CAMERA_CALIBRATION_INPUT_SANDBOXES = [
    str(Path("test/data/pedestals_LST_dark.simtel.gz").absolute()),
    str(Path("test/data/pedestals_LST_half_moon.simtel.gz").absolute()),
    str(Path("test/data/flasher_LST_dark.simtel.gz").absolute()),
    str(Path("test/data/flasher_LST_half_moon.simtel.gz").absolute()),
    str(Path("test/config/process_pedestal.yaml").absolute()),
    str(Path("test/config/process_flatfield.yaml").absolute()),
    str(Path("test/config/merge_config.yaml").absolute()),
    str(Path("test/config/pixelstats_pedestal.yaml").absolute()),
    str(Path("test/config/pixelstats_flatfield_image.yaml").absolute()),
    str(Path("test/config/pixelstats_flatfield_time.yaml").absolute()),
    str(Path("test/config/camera_calibration.yaml").absolute()),
]
CAMERA_CALIBRATION_OUTPUT_SANDBOXES = [
    "pix_stats_pedestal_image.monitoring.dl1.h5",
    "pix_stats_flatfield_image.monitoring.dl1.h5",
    "pix_stats_flatfield_time.monitoring.dl1.h5",
    "camera_calibration_lst1.mon.dl1.h5",
]
CAMERA_CALIBRATION_OUTPUT_DATA = []
CAMERA_CALIBRATION_INPUT_DATA = []


@pytest.mark.parametrize(
    (
        "expected_input_sandbox",
        "expected_input_data",
        "expected_output_sandbox",
        "expected_output_data",
    ),
    [
        (
            CAMERA_CALIBRATION_INPUT_SANDBOXES,
            CAMERA_CALIBRATION_INPUT_DATA,
            CAMERA_CALIBRATION_OUTPUT_SANDBOXES,
            CAMERA_CALIBRATION_OUTPUT_DATA,
        ),
    ],
)
def test_complex_workflow(
    mock_submit_job,
    expected_input_sandbox,
    expected_input_data,
    expected_output_sandbox,
    expected_output_data,
):
    """Test calibpipe workflow which is using
    most CWL features.
    """
    job = CWLJob(
        "tests/resources/cwl/complex_workflow/perform-camera-calibration.cwl",
        "tests/resources/cwl/complex_workflow/camera_calibration_inputs.yaml",
        cvmfs_base_path=CVMFS_BASE_PATH,
    )
    assert set(job.input_data) == set(expected_input_data)

    assert set(job.input_sandbox) == set(expected_input_sandbox)

    assert set(job.output_data) == set(expected_output_data)

    assert set(job.output_sandbox) == set(expected_output_sandbox)

    result = job.submit()
    mock_submit_job.assert_called_once_with(job)
    result_xml = ET.fromstring(result)

    input_data = get_list(result_xml, "InputData")
    assert input_data == []

    input_sandbox = get_list(result_xml, "InputSandbox")
    assert len(input_sandbox) == 13  # cwl + inputs + input_sandbox

    output_data = get_list(result_xml, "OutputData")
    assert len(output_data) == 0

    output_sandbox = get_list(result_xml, "OutputSandbox")
    assert len(output_sandbox) == 4



""" CTA specific RegisterReplica class to handle lfns starting with both the old and new VO name """

from DIRAC import S_ERROR, S_OK
from DIRAC.DataManagementSystem.Agent.RequestOperations.RegisterReplica import (
    RegisterReplica,
)
from DIRAC.DataManagementSystem.Client.DataManager import DataManager

from CTADIRAC.DataManagementSystem.Agent.RequestOperations.CTADMSRequestOperationsBase import (
    CTADMSRequestOperationsBase,
)

########################################################################


class CTARegisterReplica(RegisterReplica, CTADMSRequestOperationsBase):
    def __call__(self):
        """call me maybe"""

        self._initMonitoring()
        # # counter for failed replicas

        failedReplicas = 0
        # # catalog to use
        catalogs = self._prepareCatalogs()
        # # get waiting files
        waitingFiles = self.getWaitingFilesList()

        self._recordMonitoring("Attempted", len(waitingFiles))

        # # loop over files
        registerOperations = {}
        successReplicas = 0

        targetSE = self.operation.targetSEList[0]
        replicaTuples = [(opFile.LFN, opFile.PFN, targetSE) for opFile in waitingFiles]

        ## LA: Added for CTA-DIRAC Legacy
        first_lfn = replicaTuples[0][0]
        voName = first_lfn.split("/")[1]
        self.dm = DataManager(vo=voName)

        registerReplica = self.dm.registerReplica(replicaTuples, catalogs)

        for opFile in waitingFiles:
            lfn = opFile.LFN
            # # check results
            if not registerReplica["OK"] or lfn in registerReplica["Value"]["Failed"]:
                # There have been some errors
                self._recordMonitoring("Failed", 1)
                #        self.dataLoggingClient().addFileRecord( lfn, "RegisterReplicaFail", ','.join( catalogs )
                # if catalogs else "all catalogs", "", "RegisterReplica" )

                self._handleOpsFailure(
                    registerReplica,
                    registerOperations,
                    lfn,
                    targetSE,
                    opFile,
                    catalogs,
                    failedReplicas,
                )
            else:
                # All is OK
                if self.rmsMonitoring:
                    self._recordMonitoring("Successful", 1)
                else:
                    successReplicas += 1
                    self.log.verbose(
                        "Replica %s has been registered at %s"
                        % (lfn, ",".join(catalogs) if catalogs else "all catalogs")
                    )

                opFile.Status = "Done"

        # # if we have new replications to take place, put them at the end
        if registerOperations:
            self.log.info(f"adding {len(registerOperations)} operations to the request")
        for operation in registerOperations.values():
            self.operation._parent.addOperation(operation)

        self._commitMonitoring()

        return self._finalStatus(successReplicas, failedReplicas)

    def _prepareCatalogs(self):
        catalogs = self.operation.Catalog
        if catalogs:
            return [cat.strip() for cat in catalogs.split(",")]
        return []

    def _handleOpsFailure(
        self,
        registerReplica,
        registerOperations,
        lfn,
        targetSE,
        opFile,
        catalogs,
        failedReplicas,
    ):
        #        self.dataLoggingClient().addFileRecord( lfn, "RegisterReplicaFail", ','.join( catalogs )
        # if catalogs else "all catalogs", "", "RegisterReplica" )
        reason = registerReplica.get(
            "Message",
            registerReplica.get("Value", {}).get("Failed", {}).get(lfn, "Unknown"),
        )
        errorStr = f"failed to register LFN {lfn}: {str(reason)}"
        # FIXME: this is incompatible with the change made in the DM that we
        # ignore failures if successful in at least one catalog
        if lfn in registerReplica.get("Value", {}).get("Successful", {}) and isinstance(
            reason, dict
        ):
            # As we managed, let's create a new operation for just the remaining registration
            errorStr += " - adding registerReplica operations to request"
            self._handleFailedCatlogs(registerOperations, reason, targetSE, opFile)
            opFile.Status = "Done"
        else:
            opFile.Error = errorStr
            self._fileNotExistsInFC(reason, opFile, lfn, targetSE, catalogs)

            if opFile.Status != "Done":
                failedReplicas += 1

        self.log.warn(errorStr)

    def _fileNotExistsInFC(self, reason, opFile, lfn, targetSE, catalogs):
        catMaster = True
        if isinstance(reason, dict):
            from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

            for failedCatalog in reason:
                catMaster = catMaster and FileCatalog()._getCatalogConfigDetails(
                    failedCatalog
                ).get("Value", {}).get("Master", False)
        # If one targets explicitly a catalog and it fails or if it fails on the master catalog
        if (catalogs or catMaster) and (
            "file does not exist" in opFile.Error.lower()
            or "no such file" in opFile.Error.lower()
        ):
            # Check if the file really exists in SE, if not, consider this file registration as Done
            res = self.dm.getReplicaMetadata(lfn, targetSE)
            notExist = bool(
                "No such file" in res.get("Value", {}).get("Failed", {}).get(lfn, "")
            )
            if not notExist:
                opFile.Status = "Failed"
            else:
                opFile.Status = "Done"

    def _handleFailedCatlogs(self, registerOperations, reason, targetSE, opFile):
        for failedCatalog in reason:
            key = f"{targetSE}/{failedCatalog}"
            newOperation = self.getRegisterOperation(
                opFile,
                targetSE,
                type="RegisterReplica",
                catalog=failedCatalog,
            )
            if key not in registerOperations:
                registerOperations[key] = newOperation
            else:
                registerOperations[key].addFile(newOperation[0])

    def _finalStatus(self, successReplicas, failedReplicas):
        infoStr = ""
        if successReplicas:
            infoStr = "%d replicas successfully registered" % successReplicas
        if failedReplicas:
            infoStr += ", %d replicas failed to register" % failedReplicas
        self.log.info(f"All replicas processed {infoStr}")
        if failedReplicas:
            self.operation.Error = "some replicas failed to register"
            return S_ERROR(self.operation.Error)

        return S_OK()
